# -*- coding: utf-8 -*-
"""
The calmjs runtime collection
"""

from calmjs.runtime import SourcePackageToolchainRuntime

from calmjs.rjs.dist import extras_calmjs_methods
from calmjs.rjs.dist import source_map_methods_list
from calmjs.rjs.dist import calmjs_module_registry_methods
from calmjs.rjs.cli import create_spec
from calmjs.rjs.cli import default_toolchain
from calmjs.rjs.toolchain import STUB_MISSING_WITH_EMPTY


class RJSRuntime(SourcePackageToolchainRuntime):
    """
    Runtime for the RJSToolchain

    Example: generate a require.js artifact

    $ calmjs rjs example.package
    """

    def __init__(self, toolchain, description='r.js bundler tool', *a, **kw):
        super(RJSRuntime, self).__init__(
            cli_driver=toolchain, description=description, *a, **kw)

    def init_argparser_export_target(self, argparser):
        super(RJSRuntime, self).init_argparser_export_target(
            argparser,
            help='output filename; defaults to last ${package_name}.js',
        )

    def init_argparser_working_dir(self, argparser):
        super(RJSRuntime, self).init_argparser_working_dir(
            argparser,
            explanation=(
                'for this tool it will be used as the base directory to '
                'find source files declared for bundling; '
            ),
        )

    def init_argparser_source_registry(self, argparser):
        super(RJSRuntime, self).init_argparser_source_registry(
            argparser,
            help=(
                'comma separated list of registries to use for gathering '
                'JavaScript sources from the given Python packages; default '
                'behavior is to auto-select, enable verbose output to check '
                'to see which ones were selected'
            ),
        )

    def init_argparser(self, argparser):
        """
        Other runtimes (or users of ArgumentParser) can pass their
        subparser into here to collect the arguments here for a
        subcommand.
        """

        super(RJSRuntime, self).init_argparser(argparser)

        argparser.add_argument(
            '-e', '--empty',
            dest=STUB_MISSING_WITH_EMPTY, action='store_true',
            help="stub remaining missing modules with empty: force r.js "
                 "to ignore missing files in build directory",
        )

        argparser.add_argument(
            '--source-map-method', default='all',
            dest='source_map_method',
            choices=sorted(source_map_methods_list.keys()),
            help='the acquisition method for getting the source mappings from '
                 'the source registry for the given packages; default: all',
        )

        argparser.add_argument(
            '--source-registry-method', default='all',
            dest='source_registry_method',
            choices=sorted(calmjs_module_registry_methods.keys()),
            help='the acquisition method for getting the list of source '
                 'registries to use for the given packages; default: all',
        )

        argparser.add_argument(
            '--bundle-map-method', default='all',
            dest='bundle_map_method',
            choices=sorted(extras_calmjs_methods.keys()),
            help='the acquisition method for the bundle sources for the given '
                 'packages; default: all',
        )

        argparser.add_argument(
            '--transpile-no-indent',
            dest='transpile_no_indent', action='store_true',
            help='disable indentation of transpile sources',
        )

    def create_spec(
            self, source_package_names=(), export_target=None,
            stub_missing_with_empty=False,
            working_dir=None,
            build_dir=None,
            calmjs_module_registry_names=None,
            source_registry_method='all',
            source_map_method='all', bundle_map_method='all',
            transpile_no_indent=False,
            toolchain=None, **kwargs):
        """
        Accept all arguments, but also the explicit set of arguments
        that get passed down onto the toolchain.
        """

        # the spec takes a different set of keys as it will ultimately
        # derive the final values for the standardized spec keys.
        return create_spec(
            package_names=source_package_names,
            export_target=export_target,
            stub_missing_with_empty=stub_missing_with_empty,
            working_dir=working_dir,
            build_dir=build_dir,
            source_registry_method=source_registry_method,
            source_registries=calmjs_module_registry_names,
            source_map_method=source_map_method,
            bundle_map_method=bundle_map_method,
            transpile_no_indent=transpile_no_indent,
        )


default = RJSRuntime(default_toolchain)



# -*- coding: utf-8 -*-
"""
This provides a JavaScript "toolchain".

When I have a hammer every problem starts to look like JavaScript.

Honestly, it's a bit easier to deal with JavaScript when one treats that
as a compilation target.

How this works?

1) Write raw JS code without any UMD wrappers, but treat everything in
the file as UMD.  Remember to import everything needed using ``require``
and declare the exported things by assigning it to ``exports``.
2) Leave that file somewhere in the src directory, along with Python
code.
3) Run compile.  They will be compiled into the corresponding thing that
correlates to the Pythonic namespace identifiers.

At least this is the idea, have to see whether this idea actually end up
being sane (it won't be sane, when the entire thing was insane to begin
with).

One final thing: es6 does have modules, imports and exports done in a
different but dedicated syntax.  Though to really support them a proper
es6 compatible environment will be needed, however those are not a norm
at the moment yet, especially given the fact that browsers do not have
support for them quite just yet as of mid 2016.
"""

from __future__ import unicode_literals

import json
import logging
import sys
from os.path import dirname
from os.path import join
from os.path import exists
from os.path import isdir
from os.path import isfile
from subprocess import call

from calmjs.registry import get
from calmjs.toolchain import Toolchain
from calmjs.toolchain import CONFIG_JS_FILES
from calmjs.toolchain import EXPORT_TARGET
from calmjs.toolchain import BUILD_DIR
from calmjs.toolchain import EXPORT_MODULE_NAMES

from .utils import dict_get
from .utils import dict_key_update_overwrite_check
from .dev import rjs_advice
from .exc import RJSRuntimeError
from .exc import RJSExitError
from .registry import RJS_LOADER_PLUGIN_REGISTRY
from .registry import RJS_LOADER_PLUGIN_REGISTRY_KEY
from .registry import RJS_LOADER_PLUGIN_REGISTRY_NAME
from .requirejs import extract_all_amd_requires
from .requirejs import process_path
from .umdjs import UMD_NODE_AMD_HEADER
from .umdjs import UMD_NODE_AMD_FOOTER
from .umdjs import UMD_NODE_AMD_INDENT
from .umdjs import UMD_REQUIREJS_JSON_EXPORT_HEADER
from .umdjs import UMD_REQUIREJS_JSON_EXPORT_FOOTER

from .dist import EMPTY

logger = logging.getLogger(__name__)

_PLATFORM_SPECIFIC_RUNTIME = {
    'win32': 'r.js.cmd',
}
_DEFAULT_RUNTIME = 'r.js'

# reserved spec keys for this package
REQUIREJS_PLUGINS = 'requirejs_plugins'
STUB_MISSING_WITH_EMPTY = 'stub_missing_with_empty'


def spec_update_source_map(spec, source_map, default_source_key):
    default = dict_get(spec, default_source_key)
    for modname, source in source_map.items():
        parts = modname.split('!', 1)
        if len(parts) == 1:
            # default
            default[modname] = source
            continue

        plugin_name, arguments = parts
        plugins = dict_get(spec, REQUIREJS_PLUGINS)
        plugin = dict_get(plugins, plugin_name)
        plugin[modname] = source


def get_rjs_runtime_name(platform):
    return _PLATFORM_SPECIFIC_RUNTIME.get(platform, 'r.js')


def update_base_requirejs_config(d):
    d.update({
        'paths': {},
        'shim': {},

        'wrapShim': True,

        # other configuration options
        'optimize': "none",
        'generateSourceMaps': False,
        'normalizeDirDefines': "skip",
        'uglify': {
            'toplevel': True,
            'ascii_only': True,
            'beautify': True,
            'max_line_length': 1000,
            'defines': {
                'DEBUG': ['name', 'false']
            },
            'no_mangle': True
        },
        'uglify2': {
            'output': {
                'beautify': True
            },
            'compress': {
                'sequences': False,
                'global_defs': {
                    'DEBUG': False
                }
            },
            'warnings': True,
            'mangle': False
        },
        'useStrict': True,
        'wrap': True,
        'logLevel': 0,
    })


def _null_transpiler(spec, reader, writer):
    line = reader.readline()
    while line:
        writer.write(line)
        line = reader.readline()


def _transpile_generic_to_umd_node_amd_compat_rjs(spec, reader, writer):
    level = UMD_NODE_AMD_INDENT
    indent = '' if spec.get('transpile_no_indent') else ' ' * level
    _states = {
        'pad': 3,  # length of the header to track
    }

    def write_line(line):
        contents = line.strip()
        if _states['pad']:
            if not contents:
                writer.discard(line)
                _states['pad'] -= 1
                return
            _states['pad'] = 0
        if contents:
            writer.write_padding(indent)
        writer.write(line)

    line = reader.readline()
    if line.strip() in ("'use strict';", '"use strict";'):
        header_lines = iter(UMD_NODE_AMD_HEADER.splitlines(True))
        writer.write_padding(next(header_lines))
        writer.write_padding(next(header_lines))
        writer.write_padding(indent)
        writer.write(line)
        writer.write_padding(next(header_lines))
    else:
        writer.write_padding(UMD_NODE_AMD_HEADER)
        write_line(line)

    while line:
        line = reader.readline()
        write_line(line)

    writer.write_padding(UMD_NODE_AMD_FOOTER)


def _rjs_transpiler(spec, reader, writer):
    # ensure the reader is done from beginning
    reader.seek(0)
    line = reader.readline()
    while line and line.strip() in ('', "'use strict';", '"use strict";'):
        line = reader.readline()
    # back to the beginning
    reader.seek(0)
    if line.strip().startswith('define('):
        return _null_transpiler(spec, reader, writer)
    else:
        return _transpile_generic_to_umd_node_amd_compat_rjs(
            spec, reader, writer)


class RJSToolchain(Toolchain):
    """
    The toolchain that make use of r.js (from require.js).
    """

    rjs_bin_key = 'rjs_bin'
    rjs_bin = get_rjs_runtime_name(sys.platform)
    build_manifest_name = 'build.js'
    requirejs_config_name = 'config.js'
    node_config_name = 'node.js'

    def __init__(
            self,
            loader_plugin_registry=RJS_LOADER_PLUGIN_REGISTRY_NAME,
            *a, **kw):
        super(RJSToolchain, self).__init__(*a, **kw)
        self.loader_plugin_registry = get(loader_plugin_registry)
        self.binary = self.rjs_bin
        self._set_env_path_with_node_modules()

    def setup_transpiler(self):
        self.transpiler = _rjs_transpiler

    def build_compile_entries(self):
        return super(RJSToolchain, self).build_compile_entries() + (
            ('plugin', 'plugin', 'plugins'),
        )

    def compile_plugin(self, spec, entries):
        """
        The associated spec entry that ultimately call this should be
        prepared through this class's prepare method.
        """

        plugins_modpaths = {}
        plugins_targets = {}
        export_module_names = []

        for modname, source, target, modpath in entries:
            plugin_name, arguments = modname.split('!', 1)
            handler = spec[RJS_LOADER_PLUGIN_REGISTRY].get_record(plugin_name)
            p_pm, p_pt, m_ns = handler(
                self, spec, modname, source, target, modpath)
            _spec = locals()
            dict_key_update_overwrite_check(_spec, 'plugins_modpaths', p_pm)
            dict_key_update_overwrite_check(_spec, 'plugins_targets', p_pt)
            export_module_names.extend(m_ns)
        return plugins_modpaths, plugins_targets, export_module_names

    def modname_source_target_to_modpath(self, spec, modname, source, target):
        """
        Return 'empty:' if the source is also that, as this is the only
        way to ensure r.js won't try to bundle that location if any
        modules try to require whatever that was.  Also not raising an
        exception simply because these entries are needed to be added to
        the resulting paths.
        """

        return EMPTY if source == EMPTY else modname

    def transpile_modname_source_target(self, spec, modname, source, target):
        if source == EMPTY:
            # This is inserted by the source mapper if this item was
            # marked to be ignored for r.js, and so don't bother letting
            # parent "compile" this (which is just a simple copying)
            return
        super(RJSToolchain, self).transpile_modname_source_target(
            spec, modname, source, target)

    def prepare(self, spec):
        """
        Attempts to locate the r.js binary if not already specified.  If
        the binary file was not found, RJSRuntimeError will be raised.
        """

        loader_plugin_registry = get(spec.get(RJS_LOADER_PLUGIN_REGISTRY_KEY))
        loader_plugin_registry = spec[RJS_LOADER_PLUGIN_REGISTRY] = (
            loader_plugin_registry or self.loader_plugin_registry)

        if self.rjs_bin_key not in spec:
            which_bin = spec[self.rjs_bin_key] = (
                self.which() or self.which_with_node_modules())
            if which_bin is None:
                raise RJSRuntimeError(
                    "unable to locate '%s'" % self.binary)
            logger.debug("using '%s' as '%s'", which_bin, self.binary)
        elif not exists(spec[self.rjs_bin_key]):
            # should we check whether target can be executed?
            raise RJSRuntimeError(
                "'%s' does not exist; cannot be used as '%s' binary" % (
                    spec[self.rjs_bin_key],
                    self.rjs_bin
                )
            )

        # with requirejs, it would be nice to also build a simple config
        # that can be used from within node with the stuff in just the
        # build directory - if this wasn't already defined for some
        # reason.
        spec['requirejs_config_js'] = join(
            spec['build_dir'], self.requirejs_config_name)
        spec['node_config_js'] = join(
            spec['build_dir'], self.node_config_name)
        spec['build_manifest_path'] = join(
            spec[BUILD_DIR], self.build_manifest_name)

        if EXPORT_TARGET not in spec:
            raise RJSRuntimeError(
                "'%s' not found in spec" % EXPORT_TARGET)

        # no effect if EXPORT_TARGET already absolute.
        spec[EXPORT_TARGET] = spec[EXPORT_TARGET] = self.join_cwd(
            spec[EXPORT_TARGET])
        # Only providing the standard web one, as the node version is
        # for internal testing
        spec[CONFIG_JS_FILES] = [spec['requirejs_config_js']]

        if not isdir(dirname(spec[EXPORT_TARGET])):
            raise RJSRuntimeError(
                "'%s' will not be writable" % EXPORT_TARGET)
        logger.debug(
            "'%s' declared to be '%s'",
            EXPORT_TARGET, spec[EXPORT_TARGET]
        )

        keys = ('requirejs_config_js', 'build_manifest_path')
        matched = [k for k in keys if spec[EXPORT_TARGET] == spec[k]]

        if matched:
            raise RJSRuntimeError(
                "'%s' must not be same as '%s'" % (EXPORT_TARGET, matched[0]))

        plugin_source_map = spec['plugin_source_map'] = {}
        raw_plugins = spec.get(REQUIREJS_PLUGINS, {})
        for key, value in raw_plugins.items():
            handler = loader_plugin_registry.get_record(key)
            if handler:
                # assume handler will do the job.
                plugin_source_map.update(value)
                logger.debug("found handler for '%s' loader plugin", key)
            else:
                logger.warning(
                    "handler for '%s' loader plugin not found in registry; "
                    "as arguments associated with requirejs loader plugins "
                    "are specific, processing is disabled and the following "
                    "names will not be compiled into the target: %s",
                    key, sorted(value.keys()),
                )

        # setup own advice.
        rjs_advice(spec)

    def assemble(self, spec):
        """
        Assemble the library by compiling everything and generate the
        required files for the final bundling.
        """

        export_module_names = spec[EXPORT_MODULE_NAMES]

        # the build config is the file that will be passed to r.js for
        # building the final bundle.
        build_config = {}
        # Set up the statically defined settings.
        update_base_requirejs_config(build_config)
        build_config['shim'].update(spec.get('shim', {}))
        build_config['out'] = spec[EXPORT_TARGET]
        build_config['include'] = export_module_names

        # the requirejs config is for usage of the "built" (in this
        # case, transpiled) files, so that the import names are mapped
        # to the right location within the build_dir.  Doing this here
        # because the path handling becomes different here.
        requirejs_config = {}
        requirejs_config.update(build_config)

        # Update paths with names pointing to built files in build_dir
        # for the configuration for serving.
        requirejs_config['baseUrl'] = spec['build_dir']
        # leave as empty as this is only applicable to build
        requirejs_config['include'] = []

        # These are the configured paths
        configured_paths = {}
        # as a last resort, all targets loaded will have their soruce
        # tree inspected for module names they need; this is done so
        # that the r.js bundler will not choke when it finds missing
        # paths.
        parsed_required_paths = {}
        # the final result will be merged here.
        requirejs_config['paths'] = {}

        emptied = set()

        # correct the targets by appending a ? for the affected targets
        source_prefixes = ('transpiled', 'bundled')
        for prefix in source_prefixes:
            key = prefix + '_targets'
            modpaths_group = prefix + '_modpaths'
            for modname, target in spec[key].items():
                if spec[modpaths_group].get(modname) == EMPTY:
                    # simply omit empty exported modpaths_group.
                    emptied.add(modname)
                    continue
                if target.endswith('.js'):
                    full_target = join(spec[BUILD_DIR], *target.split('/'))
                    # requirejs loader will automatically append another
                    # .js filename extension as it doesn't know anything
                    # about the path, so to avoid this append a '?', the
                    # canonical way to tell it not to do this.
                    if isfile(full_target):
                        configured_paths[modname] = target + '?'
                        # also, do the parsing for the parsed paths
                        # this should also preemptively report potential
                        # syntax error.
                        parsed_required_paths.update({
                            modname: EMPTY for modname in (process_path(
                                full_target, extract_all_amd_requires) or [])
                        })
                        continue

                configured_paths[modname] = target

        # finally, update the config with the plugin targets, which
        # should have been correctly processed by the plugin handlers.
        configured_paths.update(spec['plugins_targets'])

        missing_modname = (
            set(parsed_required_paths) - set(configured_paths) - emptied)

        # now merge the results together and figure out the logger.
        if spec.get(STUB_MISSING_WITH_EMPTY):
            missing_logger = logger.info
            requirejs_config['paths'].update(parsed_required_paths)
            build_config['paths'].update({
                modname: parsed_required_paths[modname]
                for modname in missing_modname
            })
        else:
            # TODO adjust the message somewhat for the error case
            missing_logger = logger.error

        requirejs_config['paths'].update(configured_paths)

        # Back to the build config.  Grab only paths that have been
        # made empty and apply it to the build configuration, plus log
        # the modules.
        if missing_modname:
            missing_logger(
                'source file(s) referenced modules that are missing in the '
                'build directory: %s', ', '.join(
                    sorted(repr(m) for m in missing_modname))
            )

        prefixes = ('transpiled', 'bundled', 'plugins')
        for prefix in prefixes:
            key = prefix + '_modpaths'
            build_config['paths'].update(
                {k: v for k, v in spec[key].items() if v == EMPTY})

        # build a configuration for usage directly from nodejs (which
        # may or may not work, but a test can find out).
        nodejs_config = {}
        nodejs_config.update(build_config)
        nodejs_config['baseUrl'] = spec['build_dir']

        # write out the configuration files
        with open(spec['build_manifest_path'], 'w') as fd:
            fd.write('(\n')
            json.dump(build_config, fd, indent=4)
            fd.write('\n)')

        with open(spec['requirejs_config_js'], 'w') as fd:
            fd.write(UMD_REQUIREJS_JSON_EXPORT_HEADER)
            json.dump(requirejs_config, fd, indent=4)
            fd.write(UMD_REQUIREJS_JSON_EXPORT_FOOTER)

        with open(spec['node_config_js'], 'w') as fd:
            fd.write(UMD_REQUIREJS_JSON_EXPORT_HEADER)
            json.dump(nodejs_config, fd, indent=4)
            fd.write(UMD_REQUIREJS_JSON_EXPORT_FOOTER)

    def link(self, spec):
        """
        Basically link everything up as a bundle, as if statically
        linking everything into "binary" file.
        """

        args = (spec[self.rjs_bin_key], '-o', spec['build_manifest_path'])
        logger.info('invoking %s %s %s', *args)
        rc = call(args)
        if rc != 0:
            logger.error(
                "the spec may have contained insufficient information "
                "required for r.js to locate all dependencies it needs for "
                "the final build process."
            )
            raise RJSExitError(rc, spec[self.rjs_bin_key])



# -*- coding: utf-8 -*-
"""
Registry system for ``calmjs.rjs``.

This module provides the registries needed for the full functionality of
``calmjs.rjs``.  Only the ``LoaderPluginRegistry`` is currently provided
for the management of ``requirejs`` loader plugins for assisting with
conversion of paths to module names and target locations that are better
understood by the underlying path loader system.

Each plugin constructed will simply be a class that inherits from the
parent class ``calmjs.rjs.plugin.LoaderPluginHandler`` to faciliate type
checking, and it must accept the registry the loaded it as an argument;
this registry reference can then be used by a plugin to resolve other
sibling plugins registered to the same registry system to faciliate the
lookup of further names to build the targets needed.  For specific
implementation details, please refer to the ``calmjs.rjs.plugin``
module.

"""

from logging import getLogger

from calmjs.base import BaseRegistry
from calmjs.rjs.plugin import LoaderPluginHandler
from calmjs.rjs.utils import dict_key_update_overwrite_check

logger = getLogger(__name__)
_default_handler = LoaderPluginHandler(None)

RJS_LOADER_PLUGIN_REGISTRY_NAME = 'calmjs.rjs.loader_plugin'
RJS_LOADER_PLUGIN_REGISTRY_KEY = 'rjs_loader_plugin_registry_key'
RJS_LOADER_PLUGIN_REGISTRY = 'rjs_loader_plugin_registry'


class LoaderPluginRegistry(BaseRegistry):

    # TODO make upstream consider implement this pattern of construction
    # through a new generic base registry class of some sort.

    def _init(self):
        for entry_point in self.raw_entry_points:
            try:
                cls = entry_point.load()
            except ImportError:
                logger.warning(
                    "registry '%s' failed to load loader plugin handler for "
                    "entry point '%s'", self.registry_name, entry_point,
                )
                continue

            if not issubclass(cls, LoaderPluginHandler):
                logger.warning(
                    "entry point '%s' does not lead to a valid loader plugin "
                    "handler class", entry_point
                )
                continue

            try:
                inst = cls(self, entry_point.name)
            except Exception:
                logger.exception(
                    "the loader plugin class registered at '%s' failed "
                    "to be instantiated with the following exception",
                    entry_point,
                )
                continue

            if entry_point.name in self.records:
                old = type(self.records[entry_point.name])
                logger.warning(
                    "loader plugin handler for '%s' was already registered to "
                    "an instance of '%s:%s'; '%s' will now override this "
                    "registration",
                    entry_point.name, old.__module__, old.__name__, entry_point
                )
            self.records[entry_point.name] = inst

    def get_record(self, name):
        return self.records.get(name)

    def _mapping_to_config_paths(self, mapping, method_prefix):
        """
        Private implementation, please use the public version.

        Additional Arguments:
        method_prefix
            The prefix will be combined with '_to_config_paths' to get
            the actual resolution method from the plugin handlers.
        """

        method_name = method_prefix + '_to_config_paths'
        default_to_config_paths = getattr(_default_handler, method_name)
        paths = {}
        result = {'paths': paths}

        def map_plugin_fragment(plugin, fragment):
            handler = self.get(plugin)
            if handler is None:
                logger.warning(
                    "no handler found for loader plugin '%s', the entry "
                    "{'%s': '%s'} will be dropped from generated mapping; "
                    "this action may be fatal later.", plugin, modname, path,
                )
                return
            dict_key_update_overwrite_check(
                result, 'paths', getattr(handler, method_name)(modname, path))

        def map_default(modname, path):
            # only generate and the mapping if the path assigned to
            # modname in paths is absent or incompatible
            if paths.get(modname) != path[:-3]:
                # this is compatible.
                dict_key_update_overwrite_check(
                    result, 'paths', default_to_config_paths(modname, path))

        # bring ALL the plugin entries to the top to not affect the
        # resolution of standard modules.
        for modname, path in sorted(
                mapping.items(), key=lambda x: ('!' not in x[0], x[0], x[1])):
            plugin_fragment = modname.split('!', 1)
            if len(plugin_fragment) == 1:
                map_default(modname, path)
            else:
                map_plugin_fragment(*plugin_fragment)

        return result

    def modname_source_mapping_to_config_paths(self, mapping):
        """
        For a mapping of exported module names and the specified source
        locations, produce a mapping that is compatible for usage with
        the ``requirejs.config`` function, using the plugin loaders
        available in this module.

        Returns a dictionary with a single key ``paths`` with the value
        being the new dictionary of the mapping produced.

        Arguments:

        mapping
            A source mapping, from modname to source.
        """

        return self._mapping_to_config_paths(mapping, 'modname_source')

    def modname_target_mapping_to_config_paths(self, mapping):
        """
        For a mapping of exported module names and the specified target
        locations, produce a mapping that is compatible for usage with
        the ``requirejs.config`` function, using the plugin loaders
        available in this module.

        Returns a dictionary with a single key ``paths`` with the value
        being the new dictionary of the mapping produced.

        Arguments:

        mapping
            A target mapping, from modname to target.
        """

        return self._mapping_to_config_paths(mapping, 'modname_target')



# -*- coding: utf-8 -*-
"""
As all of the following headers and footers are directly lifted from the
UMD (Universal Module Definition) repository [1], the MIT License should
govern this file.

[1] https://github.com/umdjs/umd/

The MIT License (MIT)

Copyright (c) 2014 the UMD contributors.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from __future__ import unicode_literals


def _find_indent(s):
    t = s.splitlines()[-1]
    return len(t) - len(t.lstrip())


# Template for configurating requireJS.

UMD_REQUIREJS_JSON_EXPORT_HEADER = """\
(function() {
    'use strict';

    var requirejsOptions = (
"""

UMD_REQUIREJS_JSON_EXPORT_FOOTER = """
    );

    if (typeof exports !== 'undefined' && typeof module !== 'undefined') {
        module.exports = requirejsOptions;
    }
    if (typeof requirejs !== 'undefined' && requirejs.config) {
        requirejs.config(requirejsOptions);
    }

}());
"""

# CommonJS is basically incompatible with AMD as defined by requirejs
# and so this template had some failings if mixed in with modules that
# are defined purely in requirejs AMD syntax due to the ``var define``
# on the global scope.

UMD_COMMONJS_AMD_HEADER = """\
if (typeof exports === 'object' && typeof exports.nodeName !== 'string'
        && typeof define !== 'function') {
    var define = function (factory) {
        factory(require, exports, module);
    };
}

define(function (require, exports, module) {
"""

UMD_COMMONJS_AMD_FOOTER = """\
});
"""

# This mostly works, but certain plugins seems to be broken in some
# unspecified ways that I've failed to document earlier.

UMD_NODE_AMD_HEADER = """\
(function(define) {
    define(function (require, exports, module) {
        var exports = {};
"""

UMD_NODE_AMD_FOOTER = """
        return exports;
    });

}(
    typeof module === 'object' &&
    /* istanbul ignore next */
    module.exports &&
    /* istanbul ignore next */
    typeof define !== 'function' ?
        /* istanbul ignore next */
        function (factory) {
            module.exports = factory(require, exports, module);
        }
    :
        /* istanbul ignore next */
        define
));
"""

UMD_NODE_AMD_INDENT = _find_indent(UMD_NODE_AMD_HEADER)



# -*- coding: utf-8 -*-
"""
Helper utilities.
"""

import logging

logger = logging.getLogger(__name__)


def dict_get(d, key):
    value = d[key] = d.get(key, {})
    return value


def dict_key_update_overwrite_check(d, target, mapping):
    keys = set(d[target].keys()) & set(mapping.keys())
    for key in keys:
        if d[target][key] != mapping[key]:
            logger.warning(
                "the value of %s['%s'] is being rewritten from '%s' to '%s'; "
                "configuration may be in an invalid state.",
                target, key, d[target][key], mapping[key],
            )

    # complaints are over, finish the job.
    d[target].update(mapping)



# -*- coding: utf-8 -*-
import unittest

from calmjs.rjs import ecma


class SpecUpdateSourceMapTestCase(unittest.TestCase):

    def test_parse(self):
        text = "process.stdout.write('hello world');"
        tree = ecma.parse(text)
        parser = ecma._parser
        self.assertEqual(text, tree.to_ecma())
        ecma.parse(text)
        # the parser is not mutated.
        self.assertIs(parser, ecma._parser)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
import json
import os
import re
import sys
from os import makedirs
from os.path import exists
from os.path import join
from shutil import copytree

from pkg_resources import get_distribution

from calmjs.toolchain import Spec
from calmjs.npm import Driver
from calmjs.npm import get_npm_version
from calmjs.cli import node
from calmjs import runtime
from calmjs.registry import get as get_registry
from calmjs.utils import pretty_logging

try:
    from calmjs.dev import karma
except ImportError:  # pragma: no cover
    karma = None

from calmjs.rjs import toolchain
from calmjs.rjs import cli
from calmjs.rjs.registry import LoaderPluginRegistry

from calmjs.testing import utils
from calmjs.testing.mocks import StringIO
from calmjs.testing.mocks import WorkingSet


def skip_full_toolchain_test():  # pragma: no cover
    if get_npm_version() is None:
        return (True, 'npm not available')
    if os.environ.get('SKIP_FULL'):
        return (True, 'skipping due to SKIP_FULL environment variable')
    return (False, '')


def run_node(src, *requires):
    # cross platform node runner with require paths.
    # escape backslashes in require paths.
    return node(src % ('\n'.join(
        'require("%s");' % r.replace('\\', '\\\\') for r in requires
    )))


def cls_setup_rjs_example_package(cls):
    from calmjs import dist as calmjs_dist

    # cls.dist_dir created by setup_class_integration_environment
    cls._ep_root = join(cls.dist_dir, 'example', 'package')
    makedirs(cls._ep_root)

    test_root = join(cls._ep_root, 'tests')
    makedirs(test_root)

    math_js = join(cls._ep_root, 'math.js')
    with open(math_js, 'w') as fd:
        fd.write(
            '"use strict";\n'
            '\n'
            'exports.add = function(x, y) {\n'
            '    return x + y;\n'
            '};\n'
            '\n'
            'exports.mul = function(x, y) {\n'
            '    return x * y;\n'
            '};\n'
        )

    bad_js = join(cls._ep_root, 'bad.js')
    with open(bad_js, 'w') as fd:
        fd.write(
            '"use strict";\n'
            '\n'
            '\n'
            '\n'
            'var die = function() {\n'
            '    return notdefinedsymbol;\n'
            '};\n'
            '\n'
            'exports.die = die;\n'
        )

    # TODO derive this (line, col) from the above
    cls._bad_notdefinedsymbol = (6, 12)

    main_js = join(cls._ep_root, 'main.js')
    with open(main_js, 'w') as fd:
        fd.write(
            '"use strict";\n'
            '\n'
            'var math = require("example/package/math");\n'
            'var bad = require("example/package/bad");\n'
            '\n'
            'var main = function(trigger) {\n'
            '    console.log(math.add(1, 1));\n'
            '    console.log(math.mul(2, 2));\n'
            '    if (trigger === true) {\n'
            '        bad.die();\n'
            '    }\n'
            '};\n'
            '\n'
            'exports.main = main;\n'
        )

    # JavaScript import/module names to filesystem path.
    # Normally, these are supplied through the calmjs setuptools
    # integration framework.
    cls._example_package_map = {
        'example/package/math': math_js,
        'example/package/bad': bad_js,
        'example/package/main': main_js,
    }

    test_math_js = join(cls._ep_root, 'tests', 'test_math.js')
    with open(test_math_js, 'w') as fd:
        fd.write(
            '"use strict";\n'
            '\n'
            'var math = require("example/package/math");\n'
            '\n'
            'describe("basic math functions", function() {\n'
            '    it("addition", function() {\n'
            '        expect(math.add(3, 4)).equal(7);\n'
            '        expect(math.add(5, 6)).equal(11);\n'
            '    });\n'
            '\n'
            '    it("multiplication", function() {\n'
            '        expect(math.mul(3, 4)).equal(12);\n'
            '        expect(math.mul(5, 6)).equal(30);\n'
            '    });\n'
            '});\n'
        )

    # map for our one and only test
    cls._example_package_test_map = {
        'example/package/tests/test_math': test_math_js,
    }

    # custom loader
    loader_js = join(cls._ep_root, 'loader.js')
    with open(loader_js, 'w') as fd:
        fd.write(
            # pulled from requirejs.org's plugin documentation
            'exports.load = function (name, req, onload, config) {\n'
            '    req([name], function (value) {\n'
            '        onload(value);\n'
            '    });\n'
            '}\n'
        )

    cls._example_package_loader = {
        'example/package/loader': loader_js,
    }

    # other data
    data_js = join(cls._ep_root, 'data.js')
    with open(data_js, 'w') as fd:
        fd.write(
            '(function() { define({"results": {"item_count": 0}})}());')

    cls._example_package_data = {
        'example/package/loader!example/package/data': data_js,
    }

    # also add a proper mock distribution for this.
    utils.make_dummy_dist(None, (
        ('requires.txt', ''),
        ('calmjs_module_registry.txt', cls.registry_name),
        ('entry_points.txt', (
            '[%s]\n'
            'example.package = example.package\n'
            '[%s.tests]\n'
            'example.package = example.package.tests\n' % (
                cls.registry_name,
                cls.registry_name,
            )
        )),
    ), 'example.package', '1.0', working_dir=cls.dist_dir)

    # also include the entry_point information for this package
    utils.make_dummy_dist(None, (
        ('requires.txt', ''),
        ('entry_points.txt', (
            get_distribution('calmjs.rjs').get_metadata('entry_points.txt')
        )),
    ), 'calmjs.rjs', '0.0', working_dir=cls.dist_dir)

    # re-add it again
    calmjs_dist.default_working_set.add_entry(cls.dist_dir)
    # TODO produce package_module_map

    registry = get_registry(cls.registry_name)
    record = registry.records['example.package'] = {}
    # loader note included
    record.update(cls._example_package_map)
    registry.package_module_map['example.package'] = ['example.package']

    test_registry = get_registry(cls.registry_name + '.tests')
    test_record = test_registry.records['example.package.tests'] = {}
    test_record.update(cls._example_package_test_map)
    test_registry.package_module_map['example.package'] = [
        'example.package.tests']


@unittest.skipIf(*skip_full_toolchain_test())
class ToolchainIntegrationTestCase(unittest.TestCase):
    """
    Test out the full toolchain, involving requirejs completely.
    """

    # Ensure that requirejs is properly installed through the calmjs
    # framework and specification for this package.  This environment
    # will be reused for the duration for this test.

    @classmethod
    def setUpClass(cls):
        # nosetest will still execute setUpClass, so the test condition
        # will need to be checked here also.
        if skip_full_toolchain_test()[0]:  # pragma: no cover
            return
        cls._cwd = os.getcwd()

        utils.setup_class_install_environment(
            cls, Driver, ['calmjs.rjs'], production=False)

        # For the duration of this test, operate in the tmpdir where the
        # node_modules are available.
        os.chdir(cls._env_root)

        # This is done after the above, as the setup of the following
        # integration harness will stub out the root distribution which
        # will break the installation of real tools.
        utils.setup_class_integration_environment(cls)
        # also our test data.
        cls_setup_rjs_example_package(cls)

    @classmethod
    def tearDownClass(cls):
        # Ditto, as per above.
        if skip_full_toolchain_test()[0]:  # pragma: no cover
            return
        utils.teardown_class_integration_environment(cls)
        os.chdir(cls._cwd)
        utils.rmtree(cls._cls_tmpdir)

    def setUp(self):
        # Set up the transpiler using env_path assigned in setUpClass,
        # which installed r.js to ensure the tests will find this.
        cli.default_toolchain.env_path = self._env_path

    def tearDown(self):
        # As the manipulation is done, should set this back to its
        # default state.
        cli.default_toolchain.env_path = None

    def test_build_bundle_standard(self):
        bundle_dir = utils.mkdtemp(self)
        build_dir = utils.mkdtemp(self)
        transpile_source_map = {}
        transpile_source_map.update(self._example_package_map)
        bundle_source_map = {}
        export_target = join(bundle_dir, 'example.package.js')

        rjs = toolchain.RJSToolchain()
        spec = Spec(
            transpile_source_map=transpile_source_map,
            bundle_source_map=bundle_source_map,
            export_target=export_target,
            build_dir=build_dir,
        )
        rjs(spec)

        self.assertTrue(exists(export_target))

        # verify that the bundle works with node
        stdout, stderr = run_node(
            'var requirejs = require("requirejs");\n'
            'var define = requirejs.define;\n'
            '%s\n'
            'var main = requirejs("example/package/main");\n'
            'main.main();\n',
            export_target,
        )

        self.assertEqual(stderr, '')
        self.assertEqual(stdout, '2\n4\n')

    def test_build_bundle_no_indent(self):
        bundle_dir = utils.mkdtemp(self)
        build_dir = utils.mkdtemp(self)
        transpile_source_map = {}
        transpile_source_map.update(self._example_package_map)
        bundle_source_map = {}
        export_target = join(bundle_dir, 'example.package.js')

        rjs = toolchain.RJSToolchain()
        spec = Spec(
            transpile_source_map=transpile_source_map,
            bundle_source_map=bundle_source_map,
            export_target=export_target,
            build_dir=build_dir,
            transpile_no_indent=True,
        )
        rjs(spec)

        self.assertTrue(exists(export_target))

        stdout, stderr = run_node(
            'var requirejs = require("requirejs");\n'
            '%s\n'
            'var main = requirejs("example/package/main");\n'
            'main.main(true);\n',
            spec['node_config_js'],
        )
        self.assertIn(
            join('example', 'package', 'bad.js') + ':%d:%d' % (
                self._bad_notdefinedsymbol
            ),
            stderr
        )
        self.assertEqual(stdout, '2\n4\n')

    def test_build_bundle_with_data(self):
        bundle_dir = utils.mkdtemp(self)
        build_dir = utils.mkdtemp(self)
        transpile_source_map = {}
        transpile_source_map.update(self._example_package_map)
        # include custom loader and data
        transpile_source_map.update(self._example_package_loader)
        bundle_source_map = {}
        export_target = join(bundle_dir, 'example.package')
        requirejs_plugins = {
            'example/package/loader': self._example_package_data
        }

        custom_registry = LoaderPluginRegistry(
            'custom', _working_set=WorkingSet({
                'custom': [
                    'example/package/loader = calmjs.rjs.plugin:TextPlugin']})
        )
        rjs = toolchain.RJSToolchain()
        rjs.loader_plugin_registry = custom_registry
        spec = Spec(
            transpile_source_map=transpile_source_map,
            bundle_source_map=bundle_source_map,
            requirejs_plugins=requirejs_plugins,
            export_target=export_target,
            build_dir=build_dir,
        )

        with pretty_logging(stream=StringIO()):
            # to avoid logging the issue of mismatch map to extension
            # to stderr.
            rjs(spec)

        self.assertTrue(exists(export_target))

        # verify that the bundle works with node
        stdout, stderr = run_node(
            'var requirejs = require("requirejs");\n'
            'var define = requirejs.define;\n'
            '%s\n'
            'var result = requirejs(\n'
            '    "example/package/loader!example/package/data");\n'
            'process.stdout.write("" + result.results.item_count);\n',
            export_target,
        )

        self.assertEqual(stderr, '')
        self.assertEqual(stdout, '0')

    def test_cli_create_spec(self):
        with pretty_logging(stream=StringIO()):
            spec = cli.create_spec(
                ['site'], source_registries=(self.registry_name,))
        self.assertEqual(spec['export_target'], 'site.js')

    def test_cli_compile_all_site(self):
        # create a new working directory to install our current site
        utils.remember_cwd(self)
        working_dir = utils.mkdtemp(self)
        os.chdir(working_dir)

        # Finally, install dependencies for site in the new directory
        # normally this might be done
        # npm = Driver()
        # npm.npm_install('site', production=True)
        # However, since we have our set of fake_modules, just install
        # by copying the fake_modules dir from dist_dir into the current
        # directory.

        copytree(
            join(self.dist_dir, 'fake_modules'),
            join(working_dir, 'fake_modules'),
        )

        # Trigger the compile using the module level compile function
        spec = cli.compile_all(
            ['site'], source_registries=(self.registry_name,))
        self.assertEqual(
            spec['export_target'], join(working_dir, 'site.js'))

        # verify that the bundle works with node.  First change back to
        # directory with requirejs library installed.
        os.chdir(self._env_root)

        # The execution should then work as expected on the bundle we
        # have.
        stdout, stderr = run_node(
            'var requirejs = require("requirejs");\n'
            'var define = requirejs.define;\n'
            '%s\n'
            'var datepicker = requirejs("widget/datepicker");\n'
            'console.log(datepicker.DatePickerWidget);\n',
            spec['export_target'],
        )

        self.assertEqual(stderr, '')
        self.assertEqual(stdout, 'widget.datepicker.DatePickerWidget\n')

    def test_cli_compile_all_service(self):
        # create a new working directory to install our current site
        utils.remember_cwd(self)
        working_dir = utils.mkdtemp(self)
        os.chdir(working_dir)

        # Trigger the compile using the module level compile function,
        # but without bundling
        spec = cli.compile_all(
            ['service'], source_registries=(self.registry_name,),
            bundle_map_method='none',
        )
        self.assertEqual(
            spec['export_target'], join(working_dir, 'service.js'))

        # verify that the bundle works with node.  First change back to
        # directory with requirejs library installed.
        os.chdir(self._env_root)

        # The execution should then work as expected on the bundle we
        # have.
        stdout, stderr = run_node(
            'var requirejs = require("requirejs");\n'
            'var define = requirejs.define;\n'
            '%s\n'
            'var rpclib = requirejs("service/rpc/lib");\n'
            'console.log(rpclib.Library);\n',
            spec['export_target'],
        )

        self.assertEqual(stderr, '')
        self.assertEqual(stdout, 'service.rpc.lib.Library\n')

    def test_cli_compile_explicit_service(self):
        utils.remember_cwd(self)
        working_dir = utils.mkdtemp(self)
        os.chdir(working_dir)

        # Trigger the compile using the module level compile function,
        # but without bundling
        spec = cli.compile_all(
            ['service'], source_registries=(self.registry_name,),
            bundle_map_method='none', source_map_method='explicit',
        )
        service_js = join(working_dir, 'service.js')
        self.assertEqual(spec['export_target'], service_js)

        with open(service_js) as fd:
            self.assertIn('service/rpc/lib', fd.read())

        # build its parent js separately, too
        spec = cli.compile_all(
            ['framework'], source_registries=(self.registry_name,),
            bundle_map_method='none', source_map_method='explicit',
        )
        framework_js = join(working_dir, 'framework.js')
        self.assertEqual(spec['export_target'], framework_js)

        # verify that the bundle works with node.  First change back to
        # directory with requirejs library installed.
        os.chdir(self._env_root)

        # The execution should then work as expected if we loaded both
        # bundles.
        stdout, stderr = run_node(
            'var requirejs = require("requirejs");\n'
            'var define = requirejs.define;\n'
            '%s\n'
            'var rpclib = requirejs("service/rpc/lib");\n'
            'console.log(rpclib.Library);\n',
            framework_js,
            service_js,
        )

        self.assertEqual(stderr, '')
        self.assertEqual(stdout, 'service.rpc.lib.Library\n')

    def test_runtime_cli_help_text(self):
        utils.stub_stdouts(self)
        with self.assertRaises(SystemExit) as e:
            runtime.main(['rjs', '-h'])
        self.assertEqual(e.exception.args[0], 0)
        out = ' '.join(i.strip() for i in sys.stdout.getvalue().splitlines())
        self.assertIn(
            '--export-target EXPORT_TARGET output filename; '
            'defaults to last ${package_name}.js ', out)
        self.assertIn(
            '--working-dir WORKING_DIR the working directory; '
            'for this tool it will be used as the base directory to '
            'find source files declared for bundling; ', out)
        self.assertIn('default is current working directory', out)

    def setup_runtime_main_env(self):
        # create a new working directory to install our current site
        utils.remember_cwd(self)
        current_dir = utils.mkdtemp(self)
        target_file = join(current_dir, 'bundle.js')

        # invoke installation of "fake_modules"
        copytree(
            join(self.dist_dir, 'fake_modules'),
            join(current_dir, 'fake_modules'),
        )

        return current_dir, target_file

    def test_runtime_cli_compile_all_service(self):
        current_dir, target_file = self.setup_runtime_main_env()
        os.chdir(current_dir)

        # Invoke the thing through the main runtime
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'rjs', 'service', 'site',
                '--export-target=' + target_file,
                '--source-registry=' + self.registry_name,
            ])
        self.assertEqual(e.exception.args[0], 0)
        self.assertTrue(exists(target_file))

        # verify that the bundle works with node.  First change back to
        # directory with requirejs library installed.
        os.chdir(self._env_root)

        # The execution should then work as expected on the bundle we
        # have.
        stdout, stderr = run_node(
            'var requirejs = require("requirejs");\n'
            'var define = requirejs.define;\n'
            '%s\n'
            'var lib = requirejs("framework/lib");\n'
            'console.log(lib.Core);\n'
            'var datepicker = requirejs("widget/datepicker");\n'
            'console.log(datepicker.DatePickerWidget);\n'
            'var rpclib = requirejs("service/rpc/lib");\n'
            'console.log(rpclib.Library);\n'
            'var jquery = requirejs("jquery");\n'
            'console.log(jquery);\n'
            'var underscore = requirejs("underscore");\n'
            'console.log(underscore);\n'
            '',
            target_file
        )

        self.assertEqual(stderr, '')
        # note the names of the bundled files
        self.assertEqual(stdout, (
            'framework.lib.Core\n'
            'widget.datepicker.DatePickerWidget\n'
            'service.rpc.lib.Library\n'
            'jquery/dist/jquery.js\n'
            'underscore/underscore.js\n'
        ))

    def test_runtime_cli_compile_all_service_cwd(self):
        current_dir, target_file = self.setup_runtime_main_env()

        # Invoke the thing through the main runtime
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'rjs', 'site',
                '--export-target=' + target_file,
                '--working-dir=' + current_dir,
            ])
        self.assertEqual(e.exception.args[0], 0)
        self.assertTrue(exists(target_file))

        # verify that the bundle works with node.  First change back to
        # directory with requirejs library installed.
        os.chdir(self._env_root)

        # The execution should then work as expected on the bundle we
        # have.
        stdout, stderr = run_node(
            'var requirejs = require("requirejs");\n'
            'var define = requirejs.define;\n'
            '%s\n'
            'var lib = requirejs("framework/lib");\n'
            'console.log(lib.Core);\n'
            'var datepicker = requirejs("widget/datepicker");\n'
            'console.log(datepicker.DatePickerWidget);\n'
            'var jquery = requirejs("jquery");\n'
            'console.log(jquery);\n'
            'var underscore = requirejs("underscore");\n'
            'console.log(underscore);\n'
            '',
            target_file
        )

        self.assertEqual(stderr, '')
        # note the names of the bundled files
        self.assertEqual(stdout, (
            'framework.lib.Core\n'
            'widget.datepicker.DatePickerWidget\n'
            'jquery/dist/jquery.js\n'
            'underscore/underscore.js\n'
        ))

    def test_runtime_cli_compile_framework_simple_invocation(self):
        current_dir, target_file = self.setup_runtime_main_env()
        os.chdir(current_dir)

        # Invoke the thing through the main runtime
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'rjs', 'framework',
                '--export-target=' + target_file,
            ])
        self.assertEqual(e.exception.args[0], 0)
        self.assertTrue(exists(target_file))

        # verify that the bundle works with node.  First change back to
        # directory with requirejs library installed.
        os.chdir(self._env_root)

        # The execution should then work as expected on the bundle we
        # have.
        stdout, stderr = run_node(
            'var requirejs = require("requirejs");\n'
            'var define = requirejs.define;\n'
            '%s\n'
            'var lib = requirejs("framework/lib");\n'
            'console.log(lib.Core);\n'
            '',
            target_file
        )

        self.assertEqual(stderr, '')
        self.assertEqual(stdout, (
            'framework.lib.Core\n'
        ))

    def test_runtime_cli_compile_explicit_site(self):
        current_dir, target_file = self.setup_runtime_main_env()
        os.chdir(current_dir)

        # Invoke the thing through the main runtime
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'rjs', 'site',
                '--source-map-method=explicit',
                '--bundle-map-method=none',
                '--export-target=' + target_file,
                '--source-registry=' + self.registry_name,
            ])
        self.assertEqual(e.exception.args[0], 0)

        with open(target_file) as fd:
            contents = fd.read()

        # Since the package has no sources, and we disabled bundling of
        # sources (none works here because no code to automatically get
        # r.js to look for them), it should generate an empty bundle.
        self.assertEqual(contents, '(function () {}());')

    def test_runtime_cli_compile_explicit_registry_site(self):
        utils.stub_stdouts(self)
        current_dir, target_file = self.setup_runtime_main_env()
        os.chdir(current_dir)

        # Invoke the thing through the main runtime
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'rjs', 'site',
                '--source-registry-method=explicit',
                '--export-target=' + target_file,
            ])
        self.assertEqual(e.exception.args[0], 0)

        with open(target_file) as fd:
            contents = fd.read()

        # As the registry is NOT declared for that package, it should
        # result in nothing.
        self.assertNotIn('framework/lib', contents)
        self.assertIn(
            'no module registry declarations found using packages',
            sys.stderr.getvalue(),
        )
        self.assertIn("'site'", sys.stderr.getvalue())
        self.assertIn(
            "using acquisition method 'explicit'", sys.stderr.getvalue())

    def test_runtime_cli_bundle_method_empty(self):
        utils.stub_stdouts(self)
        current_dir, target_file = self.setup_runtime_main_env()
        os.chdir(current_dir)
        build_dir = utils.mkdtemp(self)
        widget_slim_js = join(current_dir, 'widget_slim.js')
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'rjs', 'widget',
                '--build-dir=' + build_dir,
                '--source-map-method=all',
                '--bundle-map-method=empty',
                '--export-target=' + widget_slim_js,
            ])
        self.assertEqual(e.exception.args[0], 0)
        # ensure that the bundled files are not copied
        self.assertFalse(exists(join(build_dir, 'underscore.js')))
        self.assertFalse(exists(join(build_dir, 'jquery.js')))

        with open(join(build_dir, 'build.js')) as fd:
            # strip off the header and footer
            build_js = json.loads(''.join(fd.readlines()[1:-1]))

        with open(join(build_dir, 'config.js')) as fd:
            # strip off the header and footer
            config_js = json.loads(''.join(fd.readlines()[4:-10]))

        self.assertEqual(build_js['paths'], {
            'jquery': 'empty:',
            'underscore': 'empty:',
        })
        self.assertEqual(sorted(build_js['include']), [
            'framework/lib',
            'widget/core',
            'widget/datepicker',
            'widget/richedit',
        ])

        self.assertEqual(config_js['paths'], {
            'framework/lib': 'framework/lib.js?',
            'widget/core': 'widget/core.js?',
            'widget/datepicker': 'widget/datepicker.js?',
            'widget/richedit': 'widget/richedit.js?',
        })
        self.assertEqual(config_js['include'], [])

    def test_runtime_cli_bundle_method_force_empty(self):
        utils.stub_stdouts(self)
        current_dir, target_file = self.setup_runtime_main_env()
        os.chdir(current_dir)
        build_dir = utils.mkdtemp(self)
        widget_slim_js = join(current_dir, 'widget_slim.js')
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'rjs', 'widget',
                '--build-dir=' + build_dir,
                '--empty',
                '--source-map-method=all',
                '--bundle-map-method=none',
                '--export-target=' + widget_slim_js,
            ])
        self.assertEqual(e.exception.args[0], 0)
        # ensure that the bundled files are not copied
        self.assertFalse(exists(join(build_dir, 'underscore.js')))
        self.assertFalse(exists(join(build_dir, 'jquery.js')))

        with open(join(build_dir, 'build.js')) as fd:
            # strip off the header and footer
            build_js = json.loads(''.join(fd.readlines()[1:-1]))

        with open(join(build_dir, 'config.js')) as fd:
            # strip off the header and footer
            config_js = json.loads(''.join(fd.readlines()[4:-10]))

        self.assertEqual(build_js['paths'], {
            # this is missing because no sources actually poke into it,
            # whereas the previous test it showed up as extras_calmjs
            # 'jquery': 'empty:',
            'underscore': 'empty:',
        })
        self.assertEqual(sorted(build_js['include']), [
            'framework/lib',
            'widget/core',
            'widget/datepicker',
            'widget/richedit',
        ])

        self.assertEqual(config_js['paths'], {
            'framework/lib': 'framework/lib.js?',
            'widget/core': 'widget/core.js?',
            'widget/datepicker': 'widget/datepicker.js?',
            'widget/richedit': 'widget/richedit.js?',
            # this is picked up by the source analysis when empty option
            # is appied
            'underscore': 'empty:',
        })
        self.assertEqual(config_js['include'], [])

    def test_runtime_cli_bundle_method_standard(self):
        current_dir, target_file = self.setup_runtime_main_env()
        os.chdir(current_dir)
        build_dir = utils.mkdtemp(self)
        widget_js = join(current_dir, 'widget_standard.js')
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'rjs', 'widget',
                '--build-dir=' + build_dir,
                '--source-map-method=all',
                '--bundle-map-method=all',
                '--export-target=' + widget_js,
            ])
        self.assertEqual(e.exception.args[0], 0)
        # ensure that the bundled files are copied
        self.assertTrue(exists(join(build_dir, 'underscore.js')))
        # even jquery.min.js is used, it's copied like this due to how
        # modules are renamed.
        self.assertTrue(exists(join(build_dir, 'jquery.js')))

    def test_runtime_cli_bundle_method_explicit(self):
        utils.stub_stdouts(self)
        current_dir, target_file = self.setup_runtime_main_env()
        os.chdir(current_dir)
        build_dir = utils.mkdtemp(self)
        widget_js = join(current_dir, 'widget_explicit.js')
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'rjs', 'widget',
                '--build-dir=' + build_dir,
                '--source-map-method=all',
                '--bundle-map-method=explicit',
                '--export-target=' + widget_js,
            ])
        # as the explicit option only pulled dependencies from just
        # this file, the process does not actually complete
        self.assertNotEqual(e.exception.args[0], 0)
        # ensure that the explicitly defined bundled files are copied
        self.assertFalse(exists(join(build_dir, 'underscore.js')))
        self.assertTrue(exists(join(build_dir, 'jquery.js')))

    def test_runtime_cli_compile_explicit_service_framework_widget(self):
        def run_node_with_require(*requires):
            os.chdir(self._env_root)
            return run_node(
                'var requirejs = require("requirejs");\n'
                'var define = requirejs.define;\n'
                '%s\n'
                'var lib = requirejs("framework/lib");\n'
                'console.log(lib.Core);\n'
                'var datepicker = requirejs("widget/datepicker");\n'
                'console.log(datepicker.DatePickerWidget);\n'
                'var jquery = requirejs("jquery");\n'
                'console.log(jquery);\n'
                'var underscore = requirejs("underscore");\n'
                'console.log(underscore);\n',
                *requires
            )

        def runtime_main(args, error_code=0):
            # Invoke the thing through the main runtime
            os.chdir(current_dir)
            with self.assertRaises(SystemExit) as e:
                runtime.main(args)
            self.assertEqual(e.exception.args[0], error_code)

        current_dir, target_file = self.setup_runtime_main_env()
        os.chdir(current_dir)

        # stubbing to check for a _lack_ of error message.
        utils.stub_stdouts(self)

        # Invoke the thing through the main runtime
        runtime_main([
            'rjs', 'framework', 'forms', 'service',
            '--source-map-method=explicit',
            '--export-target=' + target_file,
            '--source-registry=' + self.registry_name,
        ])
        self.assertTrue(exists(target_file))
        # no complaints about missing 'widget/*' modules
        self.assertEqual('', sys.stderr.getvalue())

        # Try running it anyway with widget missing...
        stdout, stderr = run_node_with_require(target_file)
        # This naturally will not work, so the missing module will be in
        # the error
        self.assertIn('widget', stderr)

        # try again, after building the missing widget bundle.
        widget_js = join(current_dir, 'widget.js')
        runtime_main([
            'rjs', 'widget',
            '--source-map-method=explicit',
            '--export-target=' + widget_js,
            '--source-registry=' + self.registry_name,
        ])
        # no complaints about missing 'framework/lib'
        self.assertEqual('', sys.stderr.getvalue())

        # The execution should now work if the widget bundle is loaded
        # first, and output should be as expected.
        stdout, stderr = run_node_with_require(widget_js, target_file)
        self.assertEqual(stderr, '')
        # note the names of the bundled files
        self.assertEqual(stdout, (
            'framework.lib.Core\n'
            'widget.datepicker.DatePickerWidget\n'
            'jquery/dist/jquery.min.js\n'  # from widget
            # widget_js contains this because the package 'framework'
            # declared the follow location.
            'underscore/underscore-min.js\n'
        ))

        # try again, this time the widget will NOT have underscore built
        # in as the bundles will be emptied out - however we need the
        # information in framework as the 'widget' package did NOT
        # declare the extras_calmjs for underscore so compilation will
        # fail otherwise.
        widget_slim_js = join(current_dir, 'widget_slim.js')
        runtime_main([
            'rjs', 'widget',
            '--source-map-method=all',  # using all
            '--bundle-map-method=empty',
            '--export-target=' + widget_slim_js,
            '--source-registry=' + self.registry_name,
        ])
        # no complaints about missing 'underscores' as bundle empty
        self.assertEqual('', sys.stderr.getvalue())

        # The execution should now work if the widget bundle is loaded
        # first, and output should be as expected.  This time the
        # bundles are loaded in reversed order.
        stdout, stderr = run_node_with_require(target_file, widget_slim_js)
        # Output should be as expected
        self.assertEqual(stderr, '')
        self.assertEqual(stdout, (
            'framework.lib.Core\n'
            'widget.datepicker.DatePickerWidget\n'
            'jquery/dist/jquery.min.js\n'  # from widget
            # this time, the second bundle will supply this, which has
            # the one originally sourced from the location declared by
            # the 'service' package.
            'underscore/underscore.js\n'
        ))

    def test_runtime_cli_method_none_with_empty_various(self):
        # this time, use the empty option
        utils.remember_cwd(self)
        utils.stub_stdouts(self)
        current_dir = utils.mkdtemp(self)
        export_target = join(current_dir, 'export_target.js')

        os.chdir(current_dir)
        with self.assertRaises(SystemExit) as e:
            # this should fail
            runtime.main([
                'rjs', 'service', 'site',
                '--bundle-map-method=none',
                '--export-target=' + export_target,
                '--source-registry=' + self.registry_name,
            ])

        self.assertEqual(e.exception.args[0], 1)

        os.chdir(current_dir)
        with self.assertRaises(SystemExit) as e:
            # this time, apply empty, which should automatically patch
            # the configuration
            runtime.main([
                'rjs', 'service', 'site',
                '--empty',
                '--bundle-map-method=none',
                '--export-target=' + export_target,
                '--source-registry=' + self.registry_name,
            ])

        self.assertEqual(e.exception.args[0], 0)

    def test_runtime_cli_compile_no_indent(self):
        utils.remember_cwd(self)
        target_dir = utils.mkdtemp(self)
        target_file = join(target_dir, 'bundle.js')

        # Invoke the thing through the main runtime
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'rjs', 'example.package',
                '--transpile-no-indent',
                '--export-target=' + target_file,
                '--source-registry=' + self.registry_name,
            ])
        self.assertEqual(e.exception.args[0], 0)
        self.assertTrue(exists(target_file))

        stdout, stderr = run_node(
            'var requirejs = require("requirejs");\n'
            'var define = requirejs.define;\n'
            '%s\n'
            'var main = requirejs("example/package/main");\n'
            'main.main(true);\n',
            target_file,
        )
        # The test should really test the files in the build directory,
        # but if we are doing this as an integration test, the bundle
        # should also at least maintain the same column when ran...
        patt = re.compile('%s:[0-9]+:%d' % (
            target_file.replace('\\', '\\\\'), self._bad_notdefinedsymbol[-1]))
        self.assertTrue(patt.search(stderr))
        self.assertEqual(stdout, '2\n4\n')

        # ... or, just see that the bad line is there, too.
        with open(target_file) as fd:
            bundle_source = fd.read()
        self.assertIn('\nvar die = function() {\n', bundle_source)

    def test_runtime_example_package_auto_registry(self):
        utils.stub_stdouts(self)
        current_dir = utils.mkdtemp(self)
        export_target = join(current_dir, 'example_package.js')
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'rjs', 'example.package',
                '--export-target=' + export_target,
            ])
        self.assertEqual(e.exception.args[0], 0)
        self.assertTrue(exists(export_target))


@unittest.skipIf(karma is None, 'calmjs.dev or its karma module not available')
class KarmaToolchainIntegrationTestCase(unittest.TestCase):
    """
    Test out the karma toolchain, involving requirejs completely along
    with the karma testing framework as defined by calmjs.dev
    """

    @classmethod
    def setUpClass(cls):
        cls._cwd = os.getcwd()
        utils.setup_class_install_environment(
            cls, Driver, ['calmjs.rjs', 'calmjs.dev'], production=False)

        # For the duration of this test, operate in the tmpdir where the
        # node_modules are available.
        os.chdir(cls._env_root)

        # This is done after the above, as the setup of the following
        # integration harness will stub out the root distribution which
        # will break the installation of real tools.
        utils.setup_class_integration_environment(cls)
        # also our test data.
        cls_setup_rjs_example_package(cls)

    @classmethod
    def tearDownClass(cls):
        # Ditto, as per above.
        if skip_full_toolchain_test()[0]:  # pragma: no cover
            return
        utils.teardown_class_integration_environment(cls)
        os.chdir(cls._cwd)
        utils.rmtree(cls._cls_tmpdir)

    def test_karma_test_runner_basic(self):
        utils.stub_stdouts(self)
        current_dir = utils.mkdtemp(self)
        export_target = join(current_dir, 'example_package.js')
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'karma', 'rjs', 'example.package',
                '--export-target=' + export_target,
            ])
        self.assertEqual(e.exception.args[0], 0)
        self.assertTrue(exists(export_target))

    def test_karma_test_runner_standalone_artifact(self):
        """
        What's the purpose of tests if they can't be executed any time,
        anywhere, against anything?
        """

        utils.stub_stdouts(self)
        current_dir = utils.mkdtemp(self)
        export_target = join(current_dir, 'example_package.js')
        # first, generate our bundle.
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'rjs', 'example.package', '--export-target', export_target])
        self.assertTrue(exists(export_target))

        # leverage the karma run command to run the tests provided by
        # the example.package against the resulting artifact.
        with self.assertRaises(SystemExit) as e:
            runtime.main([
                'karma', 'run',
                '--test-package', 'example.package',
                # TODO make this argument optional
                '--test-registry', self.registry_name + '.tests',
                '--artifact', export_target,
                # this is critical
                '--toolchain-package', 'calmjs.rjs',
            ])
        # tests should pass against the resultant bundle
        self.assertEqual(e.exception.args[0], 0)



# -*- coding: utf-8 -*-
import unittest

from calmjs.registry import get

from calmjs.rjs.registry import LoaderPluginRegistry
from calmjs.rjs.plugin import LoaderPluginHandler
from calmjs.rjs.plugin import TextPlugin

from calmjs.utils import pretty_logging
from calmjs.testing.mocks import StringIO
from calmjs.testing.mocks import WorkingSet


class NotPlugin(LoaderPluginRegistry):
    """yeanah"""


class BadPlugin(LoaderPluginHandler):

    def __init__(self):
        """this will not be called; missing argument"""


class DupePlugin(LoaderPluginHandler):
    """
    Dummy duplicate plugin
    """


class LoaderPluginRegistryTestCase(unittest.TestCase):

    def test_initialize_standard(self):
        # ensure that we have a proper working registry
        working_set = WorkingSet({'calmjs.rjs.loader_plugin': [
            'text = calmjs.rjs.plugin:TextPlugin',
        ]})
        registry = LoaderPluginRegistry(
            'calmjs.rjs.loader_plugin', _working_set=working_set)
        self.assertTrue(isinstance(registry.get('text'), TextPlugin))

    def test_initialize_failure_missing(self):
        working_set = WorkingSet({'calmjs.rjs.loader_plugin': [
            'not_plugin = calmjs.rjs.not_plugin:nothing',
        ]})
        # should not trigger import failure
        with pretty_logging(stream=StringIO()) as stream:
            registry = LoaderPluginRegistry(
                'calmjs.rjs.loader_plugin', _working_set=working_set)
        self.assertIsNone(registry.get('not_plugin'))
        self.assertIn(
            "registry 'calmjs.rjs.loader_plugin' failed to load loader plugin "
            "handler for entry point 'not_plugin =", stream.getvalue(),
        )

    def test_initialize_failure_not_plugin(self):
        working_set = WorkingSet({'calmjs.rjs.loader_plugin': [
            'not_plugin = calmjs.rjs.tests.test_registry:NotPlugin',
        ]})
        # should not trigger import failure
        with pretty_logging(stream=StringIO()) as stream:
            registry = LoaderPluginRegistry(
                'calmjs.rjs.loader_plugin', _working_set=working_set)
        self.assertIsNone(registry.get('not_plugin'))
        self.assertIn(
            "'not_plugin = calmjs.rjs.tests.test_registry:NotPlugin' does not "
            "lead to a valid loader plugin handler class",
            stream.getvalue()
        )

    def test_initialize_failure_bad_plugin(self):
        working_set = WorkingSet({'calmjs.rjs.loader_plugin': [
            'bad_plugin = calmjs.rjs.tests.test_registry:BadPlugin',
        ]})
        # should not trigger import failure
        with pretty_logging(stream=StringIO()) as stream:
            registry = LoaderPluginRegistry(
                'calmjs.rjs.loader_plugin', _working_set=working_set)
        self.assertIsNone(registry.get('bad_plugin'))
        self.assertIn(
            "the loader plugin class registered at 'bad_plugin = "
            "calmjs.rjs.tests.test_registry:BadPlugin' failed "
            "to be instantiated with the following exception",
            stream.getvalue()
        )

    def test_initialize_warning_dupe_plugin(self):
        # ensure that we have a proper working registry
        working_set = WorkingSet({'calmjs.rjs.loader_plugin': [
            'text = calmjs.rjs.tests.test_registry:DupePlugin',
            'text = calmjs.rjs.plugin:TextPlugin',
        ]})
        # should not trigger import failure
        with pretty_logging(stream=StringIO()) as stream:
            registry = LoaderPluginRegistry(
                'calmjs.rjs.loader_plugin', _working_set=working_set)
        self.assertIn(
            "loader plugin handler for 'text' was already registered to an "
            "instance of 'calmjs.rjs.tests.test_registry:DupePlugin'",
            stream.getvalue()
        )
        # the second one will be registered
        self.assertTrue(isinstance(registry.get('text'), TextPlugin))

    def test_initialize_integration(self):
        # Use the global set and see that the defaults are registered
        registry = get('calmjs.rjs.loader_plugin')
        text = registry.get('text')
        self.assertTrue(isinstance(text, TextPlugin))
        # should return the identity as they should all be the same.
        self.assertIs(text.registry, registry)
        self.assertIs(text.registry.get('text'), text)


class MappingConversionTestCase(unittest.TestCase):

    def setUp(self):
        working_set = WorkingSet({'calmjs.rjs.loader_plugin': [
            'text = calmjs.rjs.plugin:TextPlugin',
        ]})
        self.registry = LoaderPluginRegistry(
            'calmjs.rjs.loader_plugin', _working_set=working_set)

    def test_standard_modpath_source(self):
        result = self.registry.modname_source_mapping_to_config_paths({
            'foo/bar': '/src/foo/bar.js',
            'foo/baz': '/src/foo/baz.js',
        })
        self.assertEqual({
            'foo/bar': '/src/foo/bar.js?',
            'foo/baz': '/src/foo/baz.js?',
        }, result['paths'])
        self.assertEqual(sorted(result.keys()), ['paths'])

    def test_standard_modpath_target(self):
        result = self.registry.modname_target_mapping_to_config_paths({
            'bar/bar': '/src/bar/baz.js',
            'foo/baz': '/src/foo/baz.js',
        })
        self.assertEqual({
            'bar/bar': '/src/bar/baz.js?',
            'foo/baz': '/src/foo/baz.js?',
        }, result['paths'])

    def test_standard_modpath_plugins(self):
        result = self.registry.modname_source_mapping_to_config_paths({
            'text!foo/bar': 'text!/src/foo/bar',
            'text!example.ns/baz.txt': 'text!/src/example/ns/baz.txt',
        })
        self.assertEqual({
            'foo/bar': '/src/foo/bar',
            'example.ns/baz': '/src/example/ns/baz',
            'example.ns/baz.txt': '/src/example/ns/baz.txt',
        }, result['paths'])

    def test_standard_modpath_plugin_not_found(self):
        with pretty_logging(stream=StringIO()) as stream:
            result = self.registry.modname_source_mapping_to_config_paths({
                'nosuchplugin!foo/bar': 'nosuchplugin!/src/foo/bar',
                'text!example.ns/baz.txt': 'text!/src/example/ns/baz.txt',
            })
        err = stream.getvalue()
        self.assertEqual({
            'example.ns/baz': '/src/example/ns/baz',
            'example.ns/baz.txt': '/src/example/ns/baz.txt',
        }, result['paths'])
        self.assertIn('WARNING', err)
        self.assertIn("no handler found for loader plugin 'nosuchplugin'", err)
        self.assertIn(
            "{'nosuchplugin!foo/bar': 'nosuchplugin!/src/foo/bar'} will be "
            "dropped", err,
        )
        self.assertIn("this action may be fatal later", err)

    def test_fun_edge_case(self):
        """
        A very (un)fun edge case.

        Since the base module isn't so strict, there are situations
        where the workaround is still valid.
        """

        with pretty_logging(stream=StringIO()) as stream:
            result = self.registry.modname_source_mapping_to_config_paths({
                'foo/bar': '/src/foo/bar.js',
                'text!foo/bar.txt': '/src/foo/bar.txt',
            })
        err = stream.getvalue()
        # The mapping created by the plugin is compatible so no warnings
        # are issued.
        self.assertNotIn('WARNING', err)
        self.assertEqual({
            'foo/bar': '/src/foo/bar',
            # the exact text mapping is also present.
            'foo/bar.txt': '/src/foo/bar.txt',
        }, result['paths'])

    def test_non_working_workaround(self):
        """
        A test to show non-working workaround

        The annoyances just keeps on giving.
        """

        with pretty_logging(stream=StringIO()) as stream:
            result = self.registry.modname_source_mapping_to_config_paths({
                'foo/bar': '/src/foo/bar.js',
                'text!foo/bar.txt': '/alt/src/foo/bar.txt',
            })
        err = stream.getvalue()
        self.assertEqual({
            'foo/bar': '/src/foo/bar.js?',
            'foo/bar.txt': '/alt/src/foo/bar.txt',
        }, result['paths'])
        self.assertIn('WARNING', err)
        self.assertIn(
            "the value of paths['foo/bar'] is being rewritten from "
            "'/alt/src/foo/bar' to '/src/foo/bar.js?'; "
            "configuration may be in an invalid state.", err)

    def test_unsupported_mapping(self):
        """
        The unsupported mapping we have seen before with mismatched
        filename extensions.
        """

        with pretty_logging(stream=StringIO()) as stream:
            result = self.registry.modname_target_mapping_to_config_paths({
                'text!foo/bar.html': 'text!/alt/src/foo/bar.txt',
            })
        err = stream.getvalue()
        # mapping untouched completely, saved for removal of plugin name
        self.assertEqual({
            'foo/bar.html': '/alt/src/foo/bar.txt',
        }, result['paths'])
        self.assertIn('WARNING', err)
        self.assertIn(
            "warning triggered by mapping config.paths from modpath "
            "'text!foo/bar.html' to target 'text!/alt/src/foo/bar.txt'", err)
        self.assertIn(
            "provided modname and target has no possible workaround", err)

    def test_requirejs_is_pretty_much_completely_broken(self):
        """
        Showing how requirejs and/or requirejs-text is basically broken

        I mean, I covered how it basically can't deal with filename
        extensions correctly, so no amount of workaround can really fix
        the underlying brokenness.
        """

        with pretty_logging(stream=StringIO()) as stream:
            result = self.registry.modname_target_mapping_to_config_paths({
                'text!foo/bar.txt': 'text!/src/foo/bar.txt',
                'text!foo/bar.html': 'text!/alt/src/foo/bar.html',
            })
        err = stream.getvalue()
        # html comes before txt, since the mapping is pre-sorted in
        # alphabetical order, so txt will end up overwriting html's base
        # directory.
        self.assertEqual({
            'foo/bar': '/src/foo/bar',
            'foo/bar.txt': '/src/foo/bar.txt',
            'foo/bar.html': '/alt/src/foo/bar.html',
        }, result['paths'])
        self.assertIn('WARNING', err)
        self.assertIn("value of paths['foo/bar'] is being rewritten", err)
        self.assertIn("configuration may be in an invalid state", err)
        self.assertIn(
            "the value of paths['foo/bar'] is being rewritten from "
            "'/alt/src/foo/bar' to '/src/foo/bar'; "
            "configuration may be in an invalid state.", err)



# -*- coding: utf-8 -*-
"""
ECMA integration module.

Provides a parse function for parsing a JavaScript source text into a
source tree through the slimit module.
"""

from slimit.parser import Parser

_parser = None


# TODO name parse functions after the version of the expected input.

def parse(text):
    """
    Turn a valid JavaScript source string and turn it into a source tree
    through the Parser provided by the slimit.parser module.
    """

    global _parser
    if _parser is None:
        _parser = Parser()

    return _parser.parse(text)



# -*- coding: utf-8 -*-
import unittest

from calmjs.toolchain import Spec
from calmjs.utils import pretty_logging

from calmjs.rjs.cli import create_spec
from calmjs.rjs.cli import compile_all

from calmjs.testing.mocks import StringIO


class CliTestCase(unittest.TestCase):
    """
    Test mostly basic implementation, most of the core test will be done
    in the toolchain and/or the integration tests.
    """

    def test_create_spec_empty(self):
        with pretty_logging(stream=StringIO()) as stream:
            spec = create_spec([])

        self.assertNotIn('packages []', stream.getvalue())
        self.assertIn('no packages specified', stream.getvalue())
        self.assertTrue(isinstance(spec, Spec))
        self.assertEqual(spec['export_target'], 'calmjs.rjs.export.js')
        self.assertEqual(spec['calmjs_module_registry_names'], [])

    def test_create_spec_with_calmjs_rjs(self):
        with pretty_logging(stream=StringIO()) as stream:
            spec = create_spec(['calmjs.rjs'])
        self.assertTrue(isinstance(spec, Spec))
        self.assertEqual(spec['export_target'], 'calmjs.rjs.js')
        self.assertEqual(
            spec['calmjs_module_registry_names'], ['calmjs.module'])
        self.assertEqual(
            spec['source_package_names'], ['calmjs.rjs'])

        log = stream.getvalue()
        self.assertIn(
            "automatically picked registries ['calmjs.module'] for "
            "building source map", log,
        )

    def test_create_spec_with_calmjs_rjs_manual_source(self):
        with pretty_logging(stream=StringIO()) as stream:
            spec = create_spec(
                ['calmjs.rjs'], source_registries=['calmjs.module.tests'])
        self.assertTrue(isinstance(spec, Spec))
        self.assertEqual(spec['export_target'], 'calmjs.rjs.js')
        self.assertEqual(
            spec['calmjs_module_registry_names'], ['calmjs.module.tests'])
        self.assertEqual(
            spec['source_package_names'], ['calmjs.rjs'])

        log = stream.getvalue()
        self.assertIn(
            "using manually specified registries ['calmjs.module.tests'] for "
            "building source map", log,
        )

    def test_toolchain_empty(self):
        # dict works well enough as a null toolchain
        with pretty_logging(stream=StringIO()) as stream:
            spec = compile_all([], toolchain=dict)

        self.assertNotIn('packages []', stream.getvalue())
        self.assertIn('no packages specified', stream.getvalue())
        self.assertTrue(isinstance(spec, Spec))
        self.assertEqual(spec['export_target'], 'calmjs.rjs.export.js')

    def test_toolchain_transpile_empty(self):
        # dict works well enough as a null toolchain
        with pretty_logging(stream=StringIO()) as stream:
            spec = compile_all([], toolchain=dict, transpile_no_indent=True)

        self.assertNotIn('packages []', stream.getvalue())
        self.assertIn('no packages specified', stream.getvalue())
        self.assertTrue(isinstance(spec, Spec))
        self.assertTrue(spec['transpile_no_indent'])



# -*- coding: utf-8 -*-
"""
CalmJS RequireJS cli tools.
"""

import logging

from calmjs.toolchain import Spec
from calmjs.toolchain import BUILD_DIR
from calmjs.toolchain import CALMJS_MODULE_REGISTRY_NAMES
from calmjs.toolchain import EXPORT_TARGET
from calmjs.toolchain import SOURCE_PACKAGE_NAMES
from calmjs.rjs.toolchain import STUB_MISSING_WITH_EMPTY

from calmjs.rjs.toolchain import RJSToolchain
from calmjs.rjs.toolchain import spec_update_source_map

from calmjs.rjs.dist import generate_transpile_source_maps
from calmjs.rjs.dist import generate_bundle_source_maps
from calmjs.rjs.dist import get_calmjs_module_registry_for

default_toolchain = RJSToolchain()
logger = logging.getLogger(__name__)


def create_spec(
        package_names, export_target=None, working_dir=None, build_dir=None,
        source_registry_method='all', source_registries=None,
        source_map_method='all', bundle_map_method='all',
        stub_missing_with_empty=False,
        transpile_no_indent=False):
    """
    Produce a spec for the compilation through the RJSToolchain.

    Arguments:

    package_names
        The name of the Python package to source the dependencies from.

    export_target
        The filename for the output, can be an absolute path to a file.
        Defaults to the package_name with a '.js' suffix added in the
        working_dir.

    working_dir
        The working directory.  If the package specified any extras
        calmjs requirements (e.g. node_modules), they will be searched
        for from here.  Defaults to current working directory.

    build_dir
        The build directory.  Defaults to a temporary directory that is
        automatically removed when done.

    source_registry_method
        The acqusition method for the list of calmjs module registries
        declared for the provided package names.

        'all'
            Traverse the dependency graph for the specified package to
            acquire the declared calmjs module registries to use.
        'explicit'
            Only use the calmjs module registries declared for specified
            packages.
        'none'
            Do not acquire sources.  Useful for creating bundles of just
            the bundle sources.

    source_registries
        If the provided packages did not specify all registries or have
        declared modules in alternative but not explicitly specified
        calmjs module registries, this option can be used to pass an
        explicit list of calmjs module registries to use.  Typical use
        case is to generate tests.

    source_map_method
        The acquisition method for the source mapping for the given
        package from the source_registries specified.  Choices are
        between 'all', 'explicit' or 'none'.  Defaults to 'all'.

        'all'
            Traverse the dependency graph for the specified package to
            acquire the sources declared for each of those modules.
        'explicit'
            Only acquire the sources for the specified package.
        'none'
            Do not acquire sources.  Useful for creating bundles of just
            the bundle sources.

    bundle_map_method
        The acquisition method for the bundle sources for the given
        module.  Choices are between 'all', 'explicit' or 'none'.
        Defaults to 'all'.

        'all'
            Traverse the dependency graph for the specified package and
            acquire the declarations.
        'explicit'
            Only acquire the bundle sources declared for the specified
            package.
        'empty'
            Include all entries in a way that ensure that requirejs does
            not include them.  Useful for cases where these may be
            provided by other bundles, or for cases where only the
            declared sources are desired.
        'none'
            Do not specify any bundle files.  This only works for
            packages that have declared these as optional

        Defaults to 'all'.

    stub_missing_with_empty
        If r.js/requirejs encounters a module that it cannot find, it
        will flat out kill itself and this feature cannot be disabled
        using a simple command line level flag.  The only known way to
        get around this is sto stub every single missing file using the
        'empty:' url.  Set this to True to enable the module names found
        by the toolchain that do not lead to a valid location be stubbed
        with this value in the configuration files generated for
        consumption by RequireJS or r.js.

    transpile_no_indent
        Ensure that the transpile targets have no indents.

    """

    working_dir = working_dir if working_dir else default_toolchain.join_cwd()

    if export_target is None:
        # Take the final package name for now...
        if package_names:
            export_target = package_names[-1] + '.js'
        else:
            export_target = 'calmjs.rjs.export.js'

    spec = Spec(
        transpile_no_indent=transpile_no_indent,
    )

    if source_registries is None:
        source_registries = get_calmjs_module_registry_for(
            package_names, method=source_registry_method)
        if source_registries:
            logger.info(
                "automatically picked registries %r for building source map",
                source_registries,
            )
        elif package_names:
            logger.warning(
                "no module registry declarations found using packages %r "
                "using acquisition method '%s'",
                package_names, source_registry_method,
            )
        else:
            logger.warning('no packages specified for spec construction')
    else:
        logger.info(
            "using manually specified registries %r for building source map",
            source_registries,
        )

    spec[BUILD_DIR] = build_dir
    spec[CALMJS_MODULE_REGISTRY_NAMES] = source_registries
    spec[EXPORT_TARGET] = export_target
    spec[SOURCE_PACKAGE_NAMES] = package_names
    spec[STUB_MISSING_WITH_EMPTY] = stub_missing_with_empty

    spec_update_source_map(spec, generate_transpile_source_maps(
        package_names=package_names,
        registries=source_registries,
        method=source_map_method,
    ), 'transpile_source_map')

    spec_update_source_map(spec, generate_bundle_source_maps(
        package_names=package_names,
        working_dir=working_dir,
        method=bundle_map_method,
    ), 'bundle_source_map')

    return spec


def compile_all(
        package_names, export_target=None, working_dir=None, build_dir=None,
        source_registry_method='all', source_registries=None,
        source_map_method='all', bundle_map_method='all',
        stub_missing_with_empty=False,
        transpile_no_indent=False,
        toolchain=default_toolchain):
    """
    Invoke the r.js compiler to generate a JavaScript bundle file for a
    given Python package.  The bundle will include all the dependencies
    as specified by it and its parents.

    Arguments:

    toolchain
        The toolchain instance to use.  Default is the instance in this
        module.

    For other arguments, please refer to create_spec as they are passed
    to it.

    Naturally, this package will need all its extras calmjs declarations
    available, plus the availability of r.js, before anything can be
    done.
    """

    spec = create_spec(
        package_names=package_names,
        export_target=export_target,
        working_dir=working_dir,
        build_dir=build_dir,
        source_registry_method=source_registry_method,
        source_registries=source_registries,
        source_map_method=source_map_method,
        bundle_map_method=bundle_map_method,
        stub_missing_with_empty=stub_missing_with_empty,
        transpile_no_indent=transpile_no_indent,
    )
    toolchain(spec)
    return spec



# -*- coding: utf-8 -*-
"""
Workarounds for dumb design decisions and terrible implementation
details that the requirejs library made.
"""

# <insert expletives regarding the months of wasted time/effort here>

import logging
import re
from functools import partial

from slimit import ast

from calmjs.rjs.ecma import parse

logger = logging.getLogger(__name__)
strip_quotes = partial(re.compile('([\"\'])(.*)(\\1)').sub, '\\2')
strip_slashes = partial(re.compile(r'\\(.)').sub, '\\1')


def to_str(ast_string):
    return strip_slashes(strip_quotes(ast_string.value))


def extract_function_argument_visitor(
        node, f_name, f_argn, f_argt, visitor=None):

    if visitor is None:
        visitor = extract_function_argument_visitor

    for child in node:
        # only skimming the top function, not going in.
        if isinstance(child, ast.FunctionCall) and isinstance(
                child.identifier, ast.Identifier):
            if child.identifier.value == f_name and f_argn < len(
                    child.args) and isinstance(child.args[f_argn], f_argt):
                yield to_str(child.args[f_argn])
        else:
            # yield from visit(child)
            for value in visitor(child, f_name, f_argn, f_argt):
                yield value


def extract_function_argument(text, f_name, f_argn, f_argt=ast.String):
    """
    Extract a specific argument from a specific function name.

    Arguments:

    text
        The source text.
    f_name
        The name of the function
    f_argn
        The argument number
    f_argt
        The argument type from slimit.ast; default: slimit.ast.String
    """

    tree = parse(text)
    return list(extract_function_argument_visitor(
        tree, f_name, f_argn, f_argt))


def extract_defines(text):
    """
    For the execution of tests, there is no way to tell requirejs that
    all the modules are already available synchronously through the
    provided artifact files.  This function will extract all the define
    names which can be chucked into the requirejs.deps configuration
    section.
    """

    return extract_function_argument(text, 'define', 0)


def extract_requires(text):
    """
    The requirejs library has NO way to automatically ignore files that
    it cannot find, so we have to do it for them.  This function takes
    a source file, returns all the string literals that the source file
    use as the first argument for requires.
    """

    return extract_function_argument(text, 'require', 0)


def extract_defines_with_deps_visitor(node_map):
    """
    For the execution of tests against a pre-built artifact, there is no
    way to tell requirejs that all the modules are already available
    synchronously through the provided artifact files.  This function
    will build a dictionary with keys being the module name and value
    being a list of module names it requires synchronously.
    """

    defines = {}
    yielded = set()

    def extract_defines_visitor(node, node_name):
        f_name = 'define'
        f_argn = 0
        f_argt = ast.String

        for child in node:
            if isinstance(child, ast.FunctionCall) and isinstance(
                    child.identifier, ast.Identifier):
                if child.identifier.value == f_name and f_argn < len(
                        child.args) and isinstance(child.args[f_argn], f_argt):
                    modname = to_str(child.args[f_argn])
                    moddeps = list(extract_function_argument_visitor(
                        child, 'require', 0, ast.String))
                    if modname in defines:
                        logger.warning(
                            "module '%s' defined again in '%s'",
                            modname, node_name,
                        )
                        # don't do anything more since requirejs doesn't
                        # permit redefinition in general.
                        continue
                    defines[modname] = moddeps
            else:
                extract_defines_visitor(child, node_name)

    # first flatten it into a dependency map
    for node_name, node in node_map:
        extract_defines_visitor(node, node_name)

    # then process that map to generate the values in correct order.
    def process_defines(modname):
        if modname not in yielded:
            for modname_ in defines.get(modname, ()):
                for mn in process_defines(modname_):
                    yield mn
            if modname in defines:
                yielded.add(modname)
                yield modname
            else:
                logger.warning(
                    "module '%s' required but seems to be missing", modname)

    for modname in defines.keys():
        for mn in process_defines(modname):
            yield mn


def extract_defines_with_deps(text):
    tree = parse(text)
    return list(extract_defines_with_deps_visitor([('<text>', tree)]))


def extract_defines_with_deps_from_paths(paths):
    items = []

    def append_tree(path, text):
        items.append((path, parse(text)))

    for path in paths:
        process_path(path, partial(append_tree, path))

    return list(extract_defines_with_deps_visitor(items))


def extract_all_amd_requires(text):
    """
    Extract all require and define calls from unbundled JavaScript
    source files in both AMD and CommonJS syntax.
    """

    f_names = ('require', 'define',)
    # reserved modules
    define_wrapped = dict(enumerate(('require', 'exports', 'module',)))
    reserved = ['module']

    def visit(node):
        for child in node:
            if isinstance(child, ast.FunctionCall) and isinstance(
                    child.identifier, ast.Identifier):
                if not child.args:
                    continue

                args = child.args
                # either require or define
                standard_amd = ((
                    len(child.args) >= 2 and
                    isinstance(args[0], ast.Array) and
                    isinstance(args[1], ast.FuncExpr) and
                    child.identifier.value in f_names
                ), 0)
                # only for define
                named_define = ((
                    len(child.args) >= 3 and
                    isinstance(args[0], ast.String) and
                    isinstance(args[1], ast.Array) and
                    isinstance(args[2], ast.FuncExpr) and
                    child.identifier.value == 'define'
                ), 1)

                if (isinstance(args[0], ast.String) and
                        child.identifier.value == 'require'):
                    # only yield names just from require
                    yield to_str(args[0])
                    continue

                for checks in (standard_amd, named_define):
                    cond, pos = checks
                    if not cond:
                        continue

                    for i, node in enumerate(child.args[pos]):
                        if isinstance(node, ast.String):
                            result = to_str(node)
                            if ((result not in reserved) and (
                                    result != define_wrapped.get(i))):
                                yield result

            # yield from visit(child)
            for value in visit(child):
                yield value

    tree = parse(text)
    return visit(tree)


def process_path(path, f):
    """
    Take the path and process it through one of the above functions
    """

    try:
        with open(path) as fd:
            text = fd.read()
        return f(text)
    except (OSError, IOError) as e:
        logger.error("failed to read '%s': %s: %s", path, type(e).__name__, e)
    except SyntaxError as e:
        logger.error("syntax error in '%s': %s", path, e)



# -*- coding: utf-8 -*-
"""
Integration with various tools proided by the calmjs.dev package
"""

import logging
from os.path import join
from os.path import sep

from calmjs.exc import ToolchainAbort
from calmjs.registry import get
from calmjs.toolchain import ARTIFACT_PATHS
from calmjs.toolchain import BUILD_DIR
from calmjs.toolchain import CONFIG_JS_FILES
from calmjs.toolchain import TEST_MODULE_PATHS_MAP
from calmjs.utils import json_dump
from calmjs.utils import json_dumps

try:
    from calmjs.dev.karma import BEFORE_KARMA
except ImportError:  # pragma: no cover
    # Package not available; None is the advice blackhole
    BEFORE_KARMA = None

from calmjs.rjs.registry import RJS_LOADER_PLUGIN_REGISTRY
from calmjs.rjs.registry import RJS_LOADER_PLUGIN_REGISTRY_NAME
from calmjs.rjs.requirejs import extract_defines_with_deps_from_paths
from calmjs.rjs.umdjs import UMD_REQUIREJS_JSON_EXPORT_HEADER
from calmjs.rjs.umdjs import UMD_REQUIREJS_JSON_EXPORT_FOOTER

logger = logging.getLogger(__name__)


TEST_SCRIPT_TEMPLATE = """
var deps = %s;
var tests = %s;

var start = function() {
    // The test function will then require all the tests be available as
    // karma starts.
    require(tests, window.__karma__.start);
};

requirejs.config({
    // Karma serves files from '/base'
    baseUrl: '/base',
    // ask RequireJS to load all dependencies
    deps: deps,
    // start test run, once Require.js is done
    callback: start
});

window.DEBUG = true;
"""


def rjs_advice(spec, extras=None):
    # As requirejs has specific integration requirements with karma,
    # a test runner the calmjs.dev package provides, advise that
    # runner that before its execution, special handling needs to be
    # done to correct the generated configuration file.
    spec.advise(BEFORE_KARMA, karma_requirejs, spec)


def process_artifacts(paths):
    """
    If they are provided, assuming the defined modules there will not be
    listed as a deps for loading.
    """

    # TODO figure out how to have a flag to disable this feature for use
    # cases where this is undesirable (e.g. performance reasons).
    return extract_defines_with_deps_from_paths(paths)


def karma_requirejs(spec):
    """
    An advice for the karma runtime before execution of karma that is
    needed for integrating the requirejs framework for testing into
    karma; needed when RJSToolchain was used for artifact generation.

    This advice should be registered to BEFORE_KARMA by RJSToolchain.

    This will modify the related items in spec for the generation of the
    karma.conf.js to ensure compatibility with requirejs idioms for the
    execution of the tests through karma.
    """

    # Importing this here as these modules may not be available, so to
    # avoid potential issues, import them within the scope of this
    # function; this function should never be called if the calmjs.dev
    # python package is not available for import (and the setup should
    # not add this to a valid advice).

    try:
        from calmjs.dev import karma
    except ImportError:
        logger.error(
            "package 'calmjs.dev' not available; cannot apply requirejs "
            "specific information without karma being available."
        )
        return

    required_keys = [karma.KARMA_CONFIG, BUILD_DIR]
    for key in required_keys:
        if key not in spec:
            logger.error(
                "'%s' not provided by spec; aborting configuration for karma "
                "test runner", key
            )
            raise ToolchainAbort("spec missing key '%s'" % key)

    config = spec.get(karma.KARMA_CONFIG)
    config_files = config.get('files', [])
    build_dir = spec.get(BUILD_DIR)
    plugin_registry = spec.get(RJS_LOADER_PLUGIN_REGISTRY)
    if not plugin_registry:
        logger.warning(
            'no rjs loader plugin registry provided in spec; '
            "falling back to default registry '%s'",
            RJS_LOADER_PLUGIN_REGISTRY_NAME
        )
        plugin_registry = get(RJS_LOADER_PLUGIN_REGISTRY_NAME)

    test_module_paths_map = spec.get(TEST_MODULE_PATHS_MAP, {})
    test_conf = plugin_registry.modname_target_mapping_to_config_paths(
        test_module_paths_map)

    # Ensure '/absolute' is prefixed like so to eliminate spurious error
    # messages in the test runner, simply because the requirejs plugin
    # will try to go through this mechanism to find a timestamp and fail
    # to find its expected path, triggering the unwanted messages.  This
    # naive prefixing is actually consistent for all platforms including
    # Windows...
    new_paths = {
        # however, the actual path fragments need to be split and joined
        # with the web standard '/' separator.
        k: '/absolute' + '/'.join(v.split(sep))
        for k, v in test_conf['paths'].items()
    }

    test_conf['paths'] = new_paths
    test_config_path = spec['karma_requirejs_test_config'] = join(
        build_dir, 'requirejs_test_config.js')
    with open(test_config_path, 'w') as fd:
        fd.write(UMD_REQUIREJS_JSON_EXPORT_HEADER)
        json_dump(test_conf, fd)
        fd.write(UMD_REQUIREJS_JSON_EXPORT_FOOTER)

    # Export all the module dependencies first so they get pre-loaded
    # and thus be able to be loaded synchronously by test modules.
    deps = sorted(spec.get('export_module_names', []))

    if spec.get(ARTIFACT_PATHS):
        # TODO have a flag of some sort for flagging this as optional.
        deps.extend(process_artifacts(spec.get(ARTIFACT_PATHS)))

    # Include tests separately
    tests = sorted(test_module_paths_map.keys())

    test_script_path = spec['karma_requirejs_test_script'] = join(
        build_dir, 'karma_test_init.js')
    with open(test_script_path, 'w') as fd:
        fd.write(TEST_SCRIPT_TEMPLATE % (json_dumps(deps), json_dumps(tests)))

    frameworks = ['requirejs']
    frameworks.extend(config['frameworks'])
    config['frameworks'] = frameworks
    # rebuild the files listing in specific ordering as the load order
    # matters from within a test browser spawned by karma.
    files = []
    # first include the configuration files
    files.extend(spec.get(CONFIG_JS_FILES, []))
    # then append the test configuration path
    files.append(test_config_path)
    # then the script
    files.append(test_script_path)
    # then extend the configured paths but do not auto-include them.
    files.extend({'pattern': f, 'included': False} for f in config_files)
    # update the file listing with modifications; this will be written
    # out as part of karma.conf.js by the KarmaRuntime.
    config['files'] = files



# -*- coding: utf-8 -*-
import unittest
from os.path import join
from slimit.ast import String

from calmjs.rjs import requirejs
from calmjs.utils import pretty_logging

from calmjs.testing.mocks import StringIO
from calmjs.testing.utils import mkdtemp

# an example bundle including webpack blobs and requirejs AMD blobs
artifact = """
(function () {
(function umd(root, factory) {
        if(typeof exports === 'object' && typeof module === 'object')
                module.exports = factory();
        else if(typeof define === 'function' && define.amd)
                define('lib1',[], factory);
        else if(typeof exports === 'object')
                exports["lib1"] = factory();
        else
                root["lib1"] = factory();
})(this, function() {});
;
(function(define) {
    define('lib2',['require','exports','module'],function (
        require, exports, module) {
    });

}(
    typeof module === 'object' &&
    module.exports &&
    typeof define !== 'function' ?
        function (factory) {
            module.exports = factory(require, exports, module);
        }
    :
        define
));
}());
"""

artifact_multiple1 = """
(function () {(function(define) {
    define('lib1',['require','exports','module','lib2'],function (
        require, exports, module
    ) {
        var lib2 = require('lib2');
    });
})();

define('text',['module'], function (module) {
});

define('lib2',['require','exports','module'],function () {
});

}());
"""

artifact_multiple2 = """
define('lib1',[],function () {
    require('lib2');
    require('lib3');
});

define('lib2',[],function () {
    require('lib3');
});

define('lib4',[],function () {
    require('lib1');
});

define('lib3',[],function () {
});
"""

# missing case
artifact_multiple3 = """
define('lib1',[],function () {
    require('lib4');
    require('lib2');
    require('missing');
});

define('lib2',[],function () {
    require('lib4');
    require('missing');
});

define('lib4',[],function () {
    require('missing');
});
"""

# redefinition case
artifact_multiple4 = """
define('lib1',[],function () {
    require('lib3');
});

define('lib1',[],function () {
    require('missing');
});

define('lib3',[],function () {
});
"""


commonjs_require = """
var mod1 = require('mod1');
var mod2 = require("name/mod/mod2");
var invalid = require();

// test out dynamic require calls.
var target = mod1.target;
var dynamic = require(target);
"""

requirejs_require = """
require(['some/dummy/module1', 'some/dummy/module2'], function(mod1, mod2) {
    var mod1_alt = require('some/dummy/module1');
    var mod3_alt = require('some/dummy/module3');
});

define(['defined/alternate/module', 'some.weird.module'], function(a, b) {
});

// this is a weird one
define('some test', ['require', 'module'], function(require, module) {});

// invalid code shouldn't choke the walker.
require();
define();

// test out dynamic define calls.
var target = window.target;
var dynamic = define([target], function(target_mod) {});
"""


class ToStrTestCase(unittest.TestCase):
    """
    Test that the to_str function does operate correctly on edge cases.

    Note that the LHS (encapsulated in a String) is the raw data, while
    the RHS the quote style is similar to what JavaScript also supports,
    which is also similar to what Python does.
    """

    def test_basic(self):
        self.assertEqual(requirejs.to_str(String("'hello'")), 'hello')
        self.assertEqual(requirejs.to_str(String('"hello"')), 'hello')
        # Python escaped
        self.assertEqual(requirejs.to_str(String("'hell\"o'")), 'hell"o')
        self.assertEqual(requirejs.to_str(String('"hell\'o"')), "hell'o")

    def test_backslash(self):
        # JavaScript escaped
        self.assertEqual(requirejs.to_str(String(r"'he\'llo'")), 'he\'llo')
        self.assertEqual(requirejs.to_str(String(r"'he\"llo'")), 'he\"llo')
        self.assertEqual(requirejs.to_str(String(r"'he\\llo'")), 'he\\llo')

        self.assertEqual(requirejs.to_str(String(r'"he\'llo"')), "he\'llo")
        self.assertEqual(requirejs.to_str(String(r'"he\"llo"')), "he\"llo")


class RequireJSHelperTestCase(unittest.TestCase):
    """
    Helpers to make this thing not suck like the other.
    """

    def test_extract_function_argument_basic(self):
        results = requirejs.extract_function_argument("""
        trial(1, 2, 'hello');
        trial(1, 2, "goodbye");
        """, 'trial', 2)
        self.assertEqual(results, ['hello', 'goodbye'])

    def test_extract_function_argument_mismatches(self):
        results = requirejs.extract_function_argument("""
        trial(1, 2);
        trial(1, 2, 23, 4, 5);
        """, 'trial', 2)
        self.assertEqual(results, [])

    def test_extract_function_argument_not_nested(self):
        results = requirejs.extract_function_argument("""
        (function() {
            trial(1, 2, 'hello', trial(1, 2, 'goodbye'));
            trial(1, 2, (function() { trial(1, 2, 'goodbye')})());
        })();
        """, 'trial', 2)
        self.assertEqual(results, ['hello'])

    def test_extract_function_not_sub(self):
        results = requirejs.extract_function_argument("""
        (function() {
            log('hello');
            log('');
            console.log('goodbye');
        })();
        """, 'log', 0)
        self.assertEqual(results, ['hello', ''])

    def test_extract_on_syntax_error(self):
        with self.assertRaises(SyntaxError):
            requirejs.extract_function_argument("""
            (function() {
                console.log('hello!');
                report('');
                missing_rparen(1, 2, 'hello';
            })();
            """, 'report', 0)

    def test_extract_defines(self):
        self.assertEqual(['lib1', 'lib2'], requirejs.extract_defines(artifact))

    def test_extract_defines_amd_artifact1(self):
        result = requirejs.extract_defines_with_deps(artifact_multiple1)
        # since text has no dependencies, it can be anywhere.
        result.remove('text')
        self.assertEqual(['lib2', 'lib1'], result)

    def test_extract_defines_amd_artifact2(self):
        self.assertEqual([
            'lib3', 'lib2', 'lib1', 'lib4'
        ], requirejs.extract_defines_with_deps(artifact_multiple2))

    def test_extract_defines_amd_artifact3_missing(self):
        with pretty_logging(stream=StringIO()) as stream:
            result = requirejs.extract_defines_with_deps(artifact_multiple3)
        self.assertEqual(['lib4', 'lib2', 'lib1'], result)
        s = stream.getvalue()
        self.assertIn("module 'missing' required but seems to be missing", s)
        self.assertIn("WARNING", s)

    def test_extract_defines_amd_artifact4_dupe(self):
        with pretty_logging(stream=StringIO()) as stream:
            result = requirejs.extract_defines_with_deps(artifact_multiple4)
        self.assertEqual(['lib3', 'lib1'], result)
        s = stream.getvalue()
        self.assertNotIn("'missing' required but seems to be missing", s)
        self.assertIn("module 'lib1' defined again in '<text>'", s)
        self.assertIn("WARNING", s)

    def test_extract_requires(self):
        self.assertEqual(
            ['mod1', 'name/mod/mod2'],
            requirejs.extract_requires(commonjs_require),
        )
        self.assertEqual(
            ['mod1', 'name/mod/mod2'],
            requirejs.extract_requires(commonjs_require + requirejs_require),
        )

    def test_extract_all_amd_requires(self):
        self.assertEqual([
            'defined/alternate/module',
            'some.weird.module',
            'some/dummy/module1',
            'some/dummy/module2',
            'some/dummy/module3',
        ], sorted(set(requirejs.extract_all_amd_requires(requirejs_require))))

    def test_extract_all_amd_requires_skip_reserved(self):
        src = (
            "define('some/test', ['require', 'exports', 'module'], "
            "function(require, exports, module) {});"
        )
        self.assertEqual(
            [],
            sorted(set(requirejs.extract_all_amd_requires(src)))
        )

        src = (
            "define('some/test', ['require', 'exports', 'module', 'mod1'], "
            "function(require, exports, module, mod1) {});"
        )
        self.assertEqual(
            ['mod1'],
            sorted(set(requirejs.extract_all_amd_requires(src)))
        )

        # an actual real world one with the loader plugin
        src = "define('text', ['module'], function(module) {});"
        self.assertEqual(
            [],
            sorted(set(requirejs.extract_all_amd_requires(src)))
        )

        src = "define('alt/text', ['module', 'alt'], function(module) {});"
        self.assertEqual(
            ['alt'],
            sorted(set(requirejs.extract_all_amd_requires(src)))
        )

        src = "define('alt/text', ['alt', 'module'], function(a, module) {});"
        self.assertEqual(
            ['alt'],
            sorted(set(requirejs.extract_all_amd_requires(src)))
        )

    def test_extract_read_from_file(self):
        tmpdir = mkdtemp(self)
        src_file = join(tmpdir, 'source.js')

        with open(src_file, 'w') as fd:
            fd.write(requirejs_require)

        result = requirejs.process_path(
            src_file, requirejs.extract_all_amd_requires)

        self.assertEqual([
            'defined/alternate/module',
            'some.weird.module',
            'some/dummy/module1',
            'some/dummy/module2',
            'some/dummy/module3',
        ], sorted(set(result)))

    def test_extract_read_from_file_syntax_error(self):
        tmpdir = mkdtemp(self)
        src_file = join(tmpdir, 'source.js')

        with open(src_file, 'w') as fd:
            fd.write("define([], function () { return 'blah' }")

        with pretty_logging(stream=StringIO()) as stream:
            result = requirejs.process_path(
                src_file, requirejs.extract_requires)

        self.assertIsNone(result)
        self.assertIn('syntax error', stream.getvalue())

    def test_extract_read_from_file_error(self):
        tmpdir = mkdtemp(self)
        src_file = join(tmpdir, 'source.js')

        with pretty_logging(stream=StringIO()) as stream:
            result = requirejs.process_path(
                src_file, requirejs.extract_requires)

        self.assertIsNone(result)
        self.assertIn('No such file or directory:', stream.getvalue())
        self.assertIn(src_file, stream.getvalue())



# -*- coding: utf-8 -*-
import unittest

from calmjs.utils import pretty_logging
from calmjs.rjs import utils

from calmjs.testing.mocks import StringIO


class DictGetTestCase(unittest.TestCase):
    """
    A special get that also creates keys.
    """

    def test_dict_get(self):
        items = {}
        utils.dict_get(items, 'a_key')
        self.assertEqual(items, {'a_key': {}})

        a_key = items['a_key']
        utils.dict_get(items, 'a_key')
        self.assertIs(items['a_key'], a_key)


class DictKeyGetUpdateTestCase(unittest.TestCase):
    """
    A function for updating specific dict/spec via a key, and ensure
    that any overwritten values are warned.
    """

    def test_dict_key_update_overwrite_check_standard(self):
        a = {}
        a['base_key'] = {'k1': 'v1'}
        mapping = {'k2': 'v2'}
        with pretty_logging(logger='calmjs.rjs', stream=StringIO()) as s:
            utils.dict_key_update_overwrite_check(a, 'base_key', mapping)
        self.assertEqual(s.getvalue(), '')
        self.assertEqual(a['base_key'], {'k1': 'v1', 'k2': 'v2'})

    def test_dict_key_update_overwrite_check_no_update(self):
        a = {}
        a['base_key'] = {'k1': 'v1'}
        mapping = {'k1': 'v1'}
        with pretty_logging(logger='calmjs.rjs', stream=StringIO()) as s:
            utils.dict_key_update_overwrite_check(a, 'base_key', mapping)
        self.assertEqual(s.getvalue(), '')
        self.assertEqual(a['base_key'], {'k1': 'v1'})

    def test_dict_key_update_overwrite_check_overwritten_single(self):
        a = {}
        a['base_key'] = {'k1': 'v1'}
        mapping = {'k1': 'v2'}
        with pretty_logging(logger='calmjs.rjs', stream=StringIO()) as s:
            utils.dict_key_update_overwrite_check(a, 'base_key', mapping)
        self.assertIn(
            "value of base_key['k1'] is being rewritten from 'v1' to 'v2';",
            s.getvalue())
        self.assertEqual(a['base_key'], {'k1': 'v2'})

    def test_dict_key_update_overwrite_check_overwritten_multi(self):
        a = {}
        a['base_key'] = {'k1': 'v1', 'k2': 'v2'}
        mapping = {'k1': 'v2', 'k2': 'v4'}
        with pretty_logging(logger='calmjs.rjs', stream=StringIO()) as s:
            utils.dict_key_update_overwrite_check(a, 'base_key', mapping)
        self.assertIn(
            "value of base_key['k1'] is being rewritten from 'v1' to 'v2';",
            s.getvalue())
        self.assertIn(
            "value of base_key['k2'] is being rewritten from 'v2' to 'v4';",
            s.getvalue())
        self.assertEqual(a['base_key'], {'k1': 'v2', 'k2': 'v4'})



# -*- coding: utf-8 -*-
"""
Exceptions specific for this module.
"""


class RJSRuntimeError(RuntimeError):
    """r.js runtime error"""


class RJSExitError(RuntimeError):
    """r.js exit error, for trapping exit code"""

    def __init__(self, exit_code, binary='r.js', *a):
        self.exit_code = exit_code
        if not a:
            a = ('%s terminated with exit code %d' % (binary, exit_code),)
        super(RJSExitError, self).__init__(*a)



# -*- coding: utf-8 -*-
"""
Module that links to the calmjs.dist, for use with RJSToolchain.
"""

import logging

from os import getcwd
from os.path import join
from os.path import isdir

from calmjs.registry import get
from calmjs.dist import get_extras_calmjs
from calmjs.dist import get_module_registry_dependencies
from calmjs.dist import get_module_registry_names
from calmjs.dist import flatten_extras_calmjs
from calmjs.dist import flatten_module_registry_dependencies
from calmjs.dist import flatten_module_registry_names

logger = logging.getLogger(__name__)
EMPTY = 'empty:'
_default = 'all'


# due to usage of values specific to requirejs, upstream cannot accept
# any of the following.

def identity(v):
    return v


def empty(*a, **kw):
    return EMPTY


def map_none(*a, **kw):
    return {}


def list_none(*a, **kw):
    return []


source_map_methods_list = {
    # list of tuples of (function, filter)
    'all': (
        (flatten_module_registry_dependencies, identity),
    ),
    'explicit': (
        (flatten_module_registry_dependencies, empty),
        (get_module_registry_dependencies, identity),
    ),
    'none': (
        (map_none, empty),
    ),
}

calmjs_module_registry_methods = {
    'all': flatten_module_registry_names,
    'explicit': get_module_registry_names,
    'none': list_none,
}

extras_calmjs_methods = {
    # function, joiner
    'all': (flatten_extras_calmjs, join),
    'explicit': (get_extras_calmjs, join),
    'empty': (flatten_extras_calmjs, empty),
    'none':  (map_none, empty),
}


def acquire_method(methods, key, default=_default):
    return methods.get(key, methods.get(default))


def get_calmjs_module_registry_for(package_names, method=_default):
    """
    Acquire the module registries required for the package_names.

    package_names
        The names of the Python package to generate the source maps for.
    method
        The method to acquire the dependencies for the given module
        across all the registries specified.  Choices are between 'all',
        'explicit' or 'none'.  Defaults to 'all'.

        'all'
            Traverse the dependency graph for the specified package to
            acquire the mappings declared for each of those modules.
        'explicit'
            Same as all, however all will be stubbed out using 'empty:'
            to prevent bundling.  Only the declared sources for the
            specified packages will be untouched.
        'none'
            Produce an empty source map.

        All options not on above list defaults to 'all'
    """

    registries = acquire_method(
        calmjs_module_registry_methods, method)(package_names)
    return registries


def generate_transpile_source_maps(
        package_names, registries=('calmjs.modules'), method=_default):
    """
    Invoke the module_registry_dependencies family of dist functions,
    with the specified registries, to produce the required source maps.

    Arguments:

    package_names
        The names of the Python package to generate the source maps for.
    registries
        The names of the registries to source the packages from.  If
        unspecified, pick the options declared by the provided packages.
    method
        The method to acquire the dependencies for the given module
        across all the registries specified.  Choices are between 'all',
        'explicit' or 'none'.  Defaults to 'all'.

        'all'
            Traverse the dependency graph for the specified package to
            acquire the mappings declared for each of those modules.
        'explicit'
            Same as all, however all will be stubbed out using 'empty:'
            to prevent bundling.  Only the declared sources for the
            specified packages will be untouched.
        'none'
            Produce an empty source map.

        Defaults to 'all'.
    """

    source_map_methods = acquire_method(source_map_methods_list, method)
    transpile_source_map = {}

    # source mapping functions loop first, to prevent subsequent
    # registries from providing key-value pairs via flatten that
    # overwrite key-value pairs set initially by the get method for a
    # key that has results for flatten but no results for get; in other
    # words, this ensures the get source mapping function get executed
    # last across all registeries, if needed.

    for source_f, n_filter in source_map_methods:
        for registry_name in registries:
            transpile_source_map.update(
                (k, n_filter(v)) for k, v in source_f(
                    package_names, registry_name=registry_name
                ).items()
            )

    return transpile_source_map


def generate_bundle_source_maps(
        package_names, working_dir=None, method=_default):
    """
    Acquire the bundle source maps through the calmjs registry system.

    Arguments:

    package_names
        The names of the package to acquire the sources for.
    working_dir
        The working directory.  Defaults to current working directory.
    method
        The method to acquire the bundle sources for the given module.
        Choices are between 'all', 'explicit', 'none', or 'empty'.

        'all'
            Traverse the dependency graph for the specified package and
            acquire the declarations. [default]
        'explicit'
            Only acquire the bundle sources declared for the specified
            package.
        'none'
            Produce an empty source map.  For requirejs, this means the
            default fallback behavior of loading from the base_dir (i.e.
            the build_dir) which will result in error on missing files.
            However this is left here for low level manipulation and/or
            usage.
        'empty'
            Same as all, but all paths will be replaced with 'empty:'.
            This effectively achieves the same effect as 'none', however
            in a way that should not error if the packages at hand have
            declared all the extra sources used in the extras_calmjs
            under the appropriate keys.

        Defaults to 'all'.
    """

    working_dir = working_dir if working_dir else getcwd()
    methods = acquire_method(extras_calmjs_methods, method)

    acquire_extras_calmjs, joiner = methods

    # the extras keys will be treated as valid Node.js package manager
    # subdirectories.
    valid_pkgmgr_dirs = set(get('calmjs.extras_keys').iter_records())
    extras_calmjs = acquire_extras_calmjs(package_names)
    bundle_source_map = {}

    for mgr in extras_calmjs:
        if mgr not in valid_pkgmgr_dirs:
            continue
        basedir = join(working_dir, mgr)
        if not isdir(basedir):
            if extras_calmjs[mgr]:
                logger.warning(
                    "acquired extras_calmjs needs from '%s', but working "
                    "directory '%s' does not contain it; bundling may fail.",
                    mgr, working_dir
                )
            continue  # pragma: no cover

        for k, v in extras_calmjs[mgr].items():
            bundle_source_map[k] = joiner(basedir, *(v.split('/')))

    return bundle_source_map



# -*- coding: utf-8 -*-
import unittest
from os.path import exists
from os.path import join

# This is for emulating absent calmjs.dev module
try:
    import builtins
except ImportError:  # pragma: no cover
    # python 2
    import __builtin__ as builtins

try:
    from calmjs.dev import karma
except ImportError:  # pragma: no cover
    karma = None

from calmjs.exc import ToolchainAbort
from calmjs.registry import get
from calmjs.toolchain import Spec
from calmjs.utils import pretty_logging

from calmjs.rjs.dev import karma_requirejs
from calmjs.rjs.dev import process_artifacts
from calmjs.rjs.registry import RJS_LOADER_PLUGIN_REGISTRY_NAME

from calmjs.testing.mocks import StringIO
from calmjs.testing.utils import mkdtemp
from calmjs.testing.utils import stub_item_attr_value


class ProcessArtifactsTestCase(unittest.TestCase):

    def test_process_paths(self):
        build_dir = mkdtemp(self)
        source1 = join(build_dir, 'source1.js')
        source2 = join(build_dir, 'source2.js')

        with open(source1, 'w') as fd:
            fd.write(
                "define('source1/mod1', ['require','exports','module'],"
                "function (require, exports, module) {});\n"
                "define('source1/mod2', ['require','exports','module'],"
                "function (require, exports, module) {});\n"
            )

        with pretty_logging(stream=StringIO()) as s:
            result = process_artifacts([source1, source2])

        self.assertEqual(sorted(result), ['source1/mod1', 'source1/mod2'])
        self.assertIn('failed to read', s.getvalue())
        self.assertIn(source2, s.getvalue())
        self.assertNotIn(source1, s.getvalue())

    def test_process_path_error(self):
        build_dir = mkdtemp(self)
        source1 = join(build_dir, 'source1.js')
        source2 = join(build_dir, 'source2.js')
        source3 = join(build_dir, 'source3.js')

        with open(source1, 'w') as fd:
            fd.write(
                "define('source1/mod1', ['require','exports','module'],"
                "function (require, exports, module) {});\n"
                "define('source1/mod2', ['require','exports','module'],"
                "function (require, exports, module) {});\n"
            )

        with open(source2, 'w') as fd:
            fd.write(
                "define('source2/mod1', ['require','exports','module']"
                "function (require, exports, module) {});\n"
            )

        with open(source3, 'w') as fd:
            fd.write(
                "define('source3/mod1', ['require','exports','module'],"
                "function (require, exports, module) {});\n"
            )

        with pretty_logging(stream=StringIO()) as s:
            result = process_artifacts([source1, source2, source3])
        self.assertEqual(sorted(result), [
            'source1/mod1', 'source1/mod2', 'source3/mod1'])
        self.assertIn('syntax error in', s.getvalue())
        self.assertIn(source2, s.getvalue())


class KarmaAbsentTestCase(unittest.TestCase):
    """
    Test the injection of requirejs specific idioms into the karma
    test runner setup runtime.
    """

    def test_no_calmjs_dev(self):
        __import__ = builtins.__import__

        def import_(name, *a, **kw):
            if name == 'calmjs.dev':
                raise ImportError("No module named 'calmjs.dev'")
            return __import__(name, *a, **kw)

        stub_item_attr_value(self, builtins, '__import__', import_)
        spec = Spec()

        # just to cover the fake import above
        from calmjs.toolchain import Spec as Spec_
        self.assertIs(Spec, Spec_)

        with pretty_logging(stream=StringIO()) as s:
            karma_requirejs(spec)

        self.assertNotIn('karma_config', spec)
        self.assertIn(
            "package 'calmjs.dev' not available; cannot apply requirejs",
            s.getvalue(),
        )


@unittest.skipIf(karma is None, 'calmjs.dev or its karma module not available')
class KarmaTestCase(unittest.TestCase):

    def test_karma_setup_empty(self):
        spec = Spec()
        with pretty_logging(stream=StringIO()) as s:
            with self.assertRaises(ToolchainAbort):
                karma_requirejs(spec)

        self.assertNotIn('karma_config', spec)
        self.assertIn("'karma_config' not provided by spec", s.getvalue())

    def test_karma_setup_missing_build_dir(self):
        spec = Spec(karma_config={})
        with pretty_logging(stream=StringIO()) as s:
            with self.assertRaises(ToolchainAbort):
                karma_requirejs(spec)

        self.assertNotIn('build_dir', spec)
        self.assertIn("'build_dir' not provided by spec", s.getvalue())

    def test_karma_setup_basic_empty_case(self):
        spec = Spec(
            karma_config=karma.build_base_config(),
            build_dir=mkdtemp(self),
        )

        with pretty_logging(stream=StringIO()) as s:
            karma_requirejs(spec)

        self.assertIn(
            "no rjs loader plugin registry provided in spec; "
            "falling back to default registry 'calmjs.rjs.loader_plugin'",
            s.getvalue()
        )

        # ensure the karma bits are added.
        self.assertTrue(exists(spec['karma_requirejs_test_config']))
        self.assertTrue(exists(spec['karma_requirejs_test_script']))
        self.assertIn('requirejs', spec['karma_config']['frameworks'])

    def test_karma_setup_files(self):
        karma_config = karma.build_base_config()
        karma_config['files'] = ['example/package/lib.js']
        spec = Spec(
            karma_config=karma_config,
            build_dir=mkdtemp(self),
            rjs_loader_plugin_registry=get(RJS_LOADER_PLUGIN_REGISTRY_NAME),
        )

        with pretty_logging(stream=StringIO()) as s:
            karma_requirejs(spec)

        self.assertNotIn("no rjs loader plugin registry ", s.getvalue())

        self.assertEqual(spec['karma_config']['files'], [
            spec['karma_requirejs_test_config'],
            spec['karma_requirejs_test_script'],
            {
                'pattern': 'example/package/lib.js',
                'included': False,
            },
        ])



# -*- coding: utf-8 -*-
"""
The plugin system compatibility layer for requirejs.

Given that the plugin definitions are done on top of the requirejs
module system, the plugins are actually the top level modules, with the
arguments after the ``!`` identifier being completely opaque.  This
means that if the underlying loader expects paths in a different way it
will not understand the calmjs mapping from modname to target/modpath as
laid out.

This module provides the foundation to address this issue.  Combined
with the ``loader_plugin`` registry (see ``calmjs.rjs.registry``), this
will allow explicitly defined methods to deal with path resolution, even
for custom JavaScript plugins that are shipped with Python.  As custom
registries for this can be specified, this can cater to even the most
esoteric module setup, such as custom nested plugin configurations; this
can be implemented by calling specific functions nested within each of
the handlers, or the handler could invoke the registry system (which
under the default case is accessible through its ``self.registry``
attribute) to get the required target as the call method provides both
the toolchain and the spec.

As mentioned, loader plugin handlers should be registered to the
calmjs.rjs loader_plugin registry in order for them to be used.  There
are defaults provided, which the RJSToolchain will make use of in its
standard workflow (regsistered as ``calmjs.rjs.loader_plugin``).

One final note on module layout for ``calmjs.rjs``: this module should
not import anything from the calmjs namespace.
"""

import logging
import shutil
from os import makedirs
from os.path import dirname
from os.path import exists
from os.path import join

logger = logging.getLogger(__name__)


class LoaderPluginHandler(object):
    """
    Encapsulates a loader plugin for requirejs, this provides a
    framework to deal with path mangling and/or resolution for setting
    up the paths for usage from within a requirejs environment.
    """

    def __init__(self, registry, name=None):
        """
        The registry itself will try to construct the instance and pass
        itself into the constructor; leaving this as the default will
        enable specific plugins to load further plugins should the input
        modname has more loader plugin strings.
        """

        self.registry = registry
        self.name = name

    def strip_plugin(self, value):
        """
        Strip the first plugin fragment and return just the value.
        """

        if value.startswith(self.name + '!'):
            result = value.split('!', 1)
            return result[-1].split('!', 1)[0]
        else:
            return value

    def modname_modpath_to_config_paths(self, modname, modpath):
        """
        In the default tool, this is not used, but this is left here as
        a placeholder.
        """

        return {modname: modpath}

    def modname_target_to_config_paths(self, modname, target):
        """
        For a given modnames and its target, convert them to the aliased
        locations that will be compatible with how requirejs parses the
        paths argument in its configuration json for a web environment.

        The return value must be a modname and target, which could be
        modified so they fit for usage within the paths mapping for the
        requirejs configuration.

        The default implementation simply add a '?' to the target.
        """

        return {modname: target + '?'}

    def modname_source_to_config_paths(self, modname, source):
        """
        In general, there shouldn't be a difference, however in practice
        subclasses may choose to make handling of source files somewhat
        different.  This is used by the registry system for generation
        of mapping directly from the source file provided through the
        module registry system for calmjs.
        """

        return self.modname_target_to_config_paths(modname, source)

    def __call__(self, toolchain, spec, modname, source, target, modpath):
        """
        These need to provide the actual implementation required for the
        production of the final artifact, so this will need to locate
        the resources needed for this set of arguments to function.

        Each of these should return the associated bundled_modpaths,
        bundled_targets, and the export_module_name, after the copying
        or transpilation step was done.
        """

        raise NotImplementedError


class TextPlugin(LoaderPluginHandler):
    """
    A basic text loader plugin handler; does not handle further chained
    loader plugins that have been specified in the modname, as it
    assumes everything between the first and second '!' is the target.
    """

    def __init__(self, registry, name='text'):
        # just give this a default value for ease of use.
        super(TextPlugin, self).__init__(registry, name)

    def requirejs_text_issue123(self, value):
        """
        Basically, it appears a dot (``.``) character occuring anywhere
        inside that string is treated as a filename extension, and if
        the same identical path (with the ``.``) is later requested, the
        lookup from the requirejs config will NOT be triggered.  If the
        dot character is omitted, it will work.

        For further information and the test data that support the above
        observation, please refer to the issue reported at:

        https://github.com/requirejs/text/issues/123

        Specifically note the interactions of::

            var requirejsOptions = ({"paths": {
                'nodot/target': '/srv/tmp/nodot/target',
            }});

        and

            require([
                // /srv/tmp/nodot/target.d/file
                'text!nodot/target.d/file',
            ], function() {});

        even though the filename "extension" is applied to the directory
        one level down.
        """

        if value.endswith('.'):
            logger.warning(
                "trailing '.' character for config.paths is unsupported by "
                "requirejs-text; please refer to "
                "<https://github.com/requirejs/text/issues/123>"
            )

        result = value.rsplit('.', 1)
        if len(result) < 2:
            result.append('')
        return result

    def modname_target_to_config_paths(self, modname, target):
        """
        A text loader plugin will need to have its modname stripped of
        all loader plugin specific bits, and the target be either
        stripped of its filename extension OR take the dirname as it
        just does NOT understand explicit complete filenames.
        """

        # the values provided _will_ include the `text!` portion as the
        # default toolchain generation provides this.  Here strip them
        # off from both modname and target.
        modname_s = self.strip_plugin(modname)
        target_s = self.strip_plugin(target)

        # the default, ideal result to be returned, however...
        result = {modname_s: target_s}

        # the requirejs-text plugin works incorrectly, stripping the
        # loader plugin related strings should allow the underlying
        # loader for requirejs be able to lookup the actual path from
        # the ``paths`` configured.  However, this is not the case, due
        # to its special treatment of dots.  See the method for reason
        # and the actual issue.
        modname_i123, modname_x = self.requirejs_text_issue123(modname_s)
        target_i123, target_x = self.requirejs_text_issue123(target_s)

        if not (modname_x or target_x):
            # no filename extension, we are done.
            return result

        # ugh, I hate it when I have to pretend the brokeness exported
        # out of your everyday JavaScript library is normal.

        def unsupported_warning():
            logger.warning(
                "warning triggered by mapping config.paths from "
                "modpath '%s' to target '%s'",
                modname, target
            )
            logger.warning(
                'unsupported values provided for configuration for requirejs '
                'paths in the context of the text loader plugin; provided '
                'values will result in paths mapping that will malfunction in '
                'in the context of the AMD loader provided by requirejs when '
                'these settings are applied to a configuration on the web, '
                'including usage through an http-based test runner; however, '
                'this should not affect the production of artifact through '
                'the RJSToolchain, but test failures may result'
            )
            logger.warning(
                'to silence this warning and to fully mitigate against this '
                'issue, please ensure all final filename fragments or '
                'files made available through the registry provide a file '
                'name extension for both the modpath and the associated '
                'mapped target url or file on filesystem; please refer to '
                '<https://github.com/requirejs/text/issues/123> for complete '
                'details about this issue at hand'
            )
            logger.warning(
                "If possible, potential remedy to this issue may be resolved "
                "by renaming '%s' to '%s.ext'", target_s, target_s,
            )

        if modname_x == target_x:
            if not ('/' in modname_x and '/' in target_x):
                # Standard filename extension, glad to get this one
                # quickly over with.  Include the normal one in case the
                # issue gets fixed.
                return {
                    modname_i123: target_i123,
                    modname_s: target_s,
                }
            else:
                # Well, a separator is a lot more annoying to deal with;
                # this one definitely need the warning because if the
                # directory support is fixed this will no longer work.
                # not to mention requirejs-text>=2.0.13 it may not work.
                unsupported_warning()
                logger.warning('potentially working mitigations applied')
                modname_dir, modname_junk = modname_x.split('/', 1)
                target_dir, target_junk = target_x.split('/', 1)
                modname_i123_dir = '.'.join([modname_i123, modname_dir])
                target_i123_dir = '.'.join([target_i123, target_dir])
                return {
                    modname_s: target_s,
                    modname_i123: target_i123,
                    modname_i123_dir: target_i123_dir,
                }

        # For everything else, there are too much potential variance.
        #
        # Given this mapping::
        #
        #     requirejs.config({"paths": {
        #         "org.example.ns1/file": "/src/org.example/ns1/file",
        #         "org.example.ns2/file": "/alt/org.example/ns2/file",
        #     }});
        #
        # One would expect that the following would work
        #
        #     require(['text!org.example.ns1/file'], function() {});
        #
        # Except it does not, because of how requirejs deals with file
        # name extensions as something special and doing it wrong at
        # that, which lead to this problem.  Spoiler: a given filename
        # extension does not go beyond a path separator.  For a full
        # illustration of the problem, first limit it to the first
        # example (ns1).
        #
        # Under the above workaround rules, the generated filename
        # mapping for ns1 will be {"org.example": "/src/org"}, which is
        # wrong.  However, an observer that has full understanding of
        # the rules for dealing with this kind of path mangling will see
        # that the following configuration will work:
        #
        #     requirejs.config({"paths": {
        #         'org.example': '/src/org.example',  // for ns1/file
        #     }});
        #
        # With the following require() call successfully load from the
        # url /src/org.example.ns1/file.
        #
        #     require(['text!org.example.ns1/file'], function() {});
        #
        # But now the logic have to further infer and mixing and
        # matching of all of these fragments and things get get
        # complicated real fast.  If only the libraries can directly
        # lookup the config.paths and short circuit the evaluation...
        #
        # Anyway, if both the ns1 and ns2 definitions have to be added
        # now the rules become wildly ambiguous:
        #
        #     requirejs.config({"paths": {
        #         'org.example': '/src/org.example',  // for ns1/file
        #         'org.example': '/alt/org.example',  // for ns2/file
        #     }});
        #
        # Now, what do the following two statements actually load?
        #
        #     require(['text!org.example.ns1/file'], function() {});
        #     require(['text!org.example.ns2/file'], function() {});
        #
        # On my testing (with requirejs 2.3.2 and text 2.0.15), both
        # looked up via the second mapping, as expected.  We can further
        # reduce this problem to the following mapping:
        #
        #     requirejs.config({"paths": {
        #         'text.rst': '/srv/text.html',
        #     }});
        #
        # As we had demonstrated, loading the module like so::
        #
        #     require(['text!text.rst'], function() {});
        #
        # Will simply fail, only targets suffixed with another ".ext"
        # will work.  Removing the .rst and .html from the respective
        # key and value will not convey the desired mapping.  Due to
        # this utter ambiguity we are not going to even bother doing the
        # first step.  A sane developer will quickly decide that they
        # are done playing the stupid unnecessary filename extension
        # game that requirejs invented, unlike the one who wrote this
        # huge wall of text in comment form.

        unsupported_warning()
        logger.warning(
            'provided modname and target has no possible workaround; '
            'changes to generated config paths not applied'
        )
        return result

    def __call__(self, toolchain, spec, modname, source, target, modpath):

        # the write target, however, is very different from the config
        # target given the issue #123 workaround applied.
        stripped_target = self.strip_plugin(target)
        copy_target = join(spec['build_dir'], stripped_target)
        if not exists(dirname(copy_target)):
            makedirs(dirname(copy_target))
        shutil.copy(source, copy_target)

        bundled_modpaths = {modname: modpath}
        bundled_targets = self.modname_target_to_config_paths(modname, target)
        export_module_names = [modname]
        return bundled_modpaths, bundled_targets, export_module_names



from setuptools import setup
from setuptools import find_packages

version = '1.0.0'

classifiers = """
Development Status :: 5 - Production/Stable
Environment :: Console
Intended Audience :: Developers
License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)
Operating System :: OS Independent
Programming Language :: JavaScript
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3.3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
""".strip().splitlines()

package_json = {
    "dependencies": {
        "requirejs": "~2.1.17",
        # This should be provided by some templating library that
        # require text templates through the same import mechanism.
        # "requirejs-text": "~2.0.12",
    },
    "devDependencies": {
        "karma-requirejs": "~0.2.2",
    },
}


long_description = (
    open('README.rst').read()
    + '\n' +
    open('CHANGES.rst').read()
    + '\n')

setup(
    name='calmjs.rjs',
    version=version,
    description=(
        "Package for the integration of RequireJS into a Python "
        "environment via the Calmjs framework, to provide a reproducible "
        "workflow for the generation of deployable artifacts from "
        "JavaScript source code provided by Python packages in "
        "conjunction with standard JavaScript or Node.js packages sourced "
        "from npm or other similar package repositories."
    ),
    long_description=long_description,
    # Get more strings from
    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=classifiers,
    keywords='',
    author='Tommy Yu',
    author_email='tommy.yu@auckland.ac.nz',
    url='https://github.com/calmjs/calmjs.rjs',
    license='gpl',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['calmjs'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'calmjs>=2.0.0,<3',
        'slimit>=0.8.0',
    ],
    extras_require={
        'dev': [
            'calmjs.dev>=1.0.0,<2',
        ],
    },
    entry_points={
        # If to be unleashed as a standalone tool.
        # 'console_scripts': [
        #     'rjs = calmjs.rjs.runtime:default',
        # ],
        'calmjs.registry': [
            'calmjs.rjs.loader_plugin'
            ' = calmjs.rjs.registry:LoaderPluginRegistry',
        ],
        'calmjs.rjs.loader_plugin': [
            'text = calmjs.rjs.plugin:TextPlugin',
        ],
        'calmjs.runtime': [
            'rjs = calmjs.rjs.runtime:default',
        ],
        'calmjs.toolchain.advice': [
            'calmjs.dev.toolchain:KarmaToolchain = calmjs.rjs.dev:rjs_advice',
        ],
    },
    package_json=package_json,
    calmjs_module_registry=['calmjs.module'],
    test_suite="calmjs.rjs.tests.make_suite",
)



# -*- coding: utf-8 -*-
import unittest
from os.path import join
from calmjs.utils import pretty_logging
from calmjs.rjs import dist

from calmjs.testing import utils
from calmjs.testing.mocks import StringIO


class BaseDistTestCase(unittest.TestCase):
    """
    Test out dist functions
    """

    def test_acquire_method(self):
        foo = object()
        bar = object()
        r = dist.acquire_method({'foo': foo, 'bar': bar}, 'foo', default='bar')
        self.assertIs(r, foo)
        r = dist.acquire_method({'foo': foo, 'bar': bar}, 'wat', default='bar')
        self.assertIs(r, bar)


class DistIntegrationTestCase(unittest.TestCase):
    """
    A number of integration tests, using mocked up data created with the
    calmjs.testing helpers.
    """

    @classmethod
    def setUpClass(cls):
        utils.setup_class_integration_environment(cls)

    @classmethod
    def tearDownClass(cls):
        utils.teardown_class_integration_environment(cls)

    def test_generate_transpile_source_maps_none(self):
        mapping = dist.generate_transpile_source_maps(
            ['site'], registries=(self.registry_name,), method='none')
        self.assertEqual(sorted(mapping.keys()), [])

    def test_generate_transpile_source_maps_explicit_registry(self):
        mapping = dist.generate_transpile_source_maps(
            ['site'], registries=(self.registry_name,))
        self.assertEqual(sorted(mapping.keys()), [
            'forms/ui', 'framework/lib', 'widget/core', 'widget/datepicker',
            'widget/richedit',
        ])

    def test_generate_transpile_source_maps_explicit_registry_auto(self):
        # explicitly use the package only for source
        mapping = dist.generate_transpile_source_maps(
            ['site'], registries=(self.registry_name,), method='explicit')
        # all keys should be present.
        self.assertEqual(sorted(mapping.keys()), [
            'forms/ui', 'framework/lib', 'widget/core', 'widget/datepicker',
            'widget/richedit',
        ])
        self.assertEqual(
            sorted(k for k in mapping.keys() if mapping[k] != dist.EMPTY),
            [],
        )

        mapping = dist.generate_transpile_source_maps(
            ['forms'], registries=(self.registry_name,), method='explicit')
        self.assertEqual(
            sorted(k for k in mapping.keys() if mapping[k] != dist.EMPTY),
            ['forms/ui']
        )

    def test_get_calmjs_module_registry_for_site_no_registry(self):
        # since site doesn't actually define an explicit registry that
        # it needs.
        self.assertEqual(
            dist.get_calmjs_module_registry_for(['site'], method='explicit'),
            [],
        )

        self.assertEqual(
            dist.get_calmjs_module_registry_for(['site'], method='all'),
            [self.registry_name],
        )

        self.assertEqual(
            dist.get_calmjs_module_registry_for(['site'], method='none'),
            [],
        )

    def test_get_calmjs_module_registry_for_explicit_get(self):
        self.assertEqual(
            dist.get_calmjs_module_registry_for(
                ['calmjs.simulated'], method='explicit'),
            [self.registry_name],
        )

        self.assertEqual(
            dist.get_calmjs_module_registry_for(['forms'], method='auto'),
            [self.registry_name],
        )

    def test_generate_transpile_source_maps_site_explicit_method(self):
        mapping = dist.generate_transpile_source_maps(
            ['site'], registries=(self.registry_name,), method='explicit')
        # it doesn't remove this, but only mark it as empty as the
        # underlying tool will fail otherwise if any require statements
        # having that as its argument will result in r.js failing to do
        # anything.
        self.assertEqual(sorted(mapping.keys()), [
            'forms/ui', 'framework/lib', 'widget/core', 'widget/datepicker',
            'widget/richedit',
        ])

        self.assertEqual(sorted(
            [k for k, v in mapping.items() if v == 'empty:']
        ), [
            'forms/ui', 'framework/lib', 'widget/core', 'widget/datepicker',
            'widget/richedit',
        ])

    def test_generate_transpile_source_maps_service_default(self):
        mapping = dist.generate_transpile_source_maps(
            ['service'], registries=(self.registry_name,))
        self.assertEqual(sorted(mapping.keys()), [
            'framework/lib', 'service/endpoint', 'service/rpc/lib',
        ])

    def test_generate_transpile_source_maps_service_explicit(self):
        mapping = dist.generate_transpile_source_maps(
            ['service'], registries=(self.registry_name,), method='explicit')
        self.assertEqual(sorted(mapping.keys()), [
            'framework/lib', 'service/endpoint', 'service/rpc/lib',
        ])
        self.assertEqual(mapping['framework/lib'], 'empty:')
        self.assertNotEqual(mapping['service/endpoint'], 'empty:')
        self.assertNotEqual(mapping['service/rpc/lib'], 'empty:')

    def test_generate_bundle_source_maps_none(self):
        mapping = dist.generate_bundle_source_maps(
            ['site'], method='none')
        self.assertEqual(sorted(mapping.keys()), [])

    def test_generate_bundle_source_maps_bad_dir(self):
        bad_dir = utils.mkdtemp(self)
        with pretty_logging(stream=StringIO()) as log:
            mapping = dist.generate_bundle_source_maps(
                ['service'], bad_dir)
        self.assertEqual(sorted(mapping.keys()), [])
        self.assertIn('fake_modules', log.getvalue())

    def test_generate_bundle_source_maps_site_default(self):
        mapping = dist.generate_bundle_source_maps(
            ['site'], self.dist_dir)
        self.assertEqual(sorted(mapping.keys()), ['jquery', 'underscore'])
        self.assertTrue(mapping['jquery'].endswith(
            join('fake_modules', 'jquery', 'dist', 'jquery.js')))
        self.assertTrue(mapping['underscore'].endswith(
            join('fake_modules', 'underscore', 'underscore.js')))

    def test_generate_bundle_source_maps_default(self):
        mapping = dist.generate_bundle_source_maps(
            ['framework'], self.dist_dir)
        self.assertEqual(sorted(mapping.keys()), [
            'jquery', 'underscore',
        ])
        self.assertIn(
            (join('underscore', 'underscore-min.js')), mapping['underscore'])
        mapping = dist.generate_bundle_source_maps(
            ['service'], self.dist_dir)
        self.assertEqual(sorted(mapping.keys()), [
            'jquery', 'underscore',
        ])
        self.assertIn(
            (join('underscore', 'underscore.js')), mapping['underscore'])
        self.assertIn('jquery', mapping['jquery'])

    def test_generate_bundle_source_maps_service_explicit(self):
        mapping = dist.generate_bundle_source_maps(
            ['service'], self.dist_dir, method='explicit')
        self.assertEqual(sorted(mapping.keys()), ['underscore'])
        self.assertIn(
            (join('underscore', 'underscore.js')), mapping['underscore'])

    def test_generate_bundle_source_maps_service_empty(self):
        mapping = dist.generate_bundle_source_maps(
            ['service'], self.dist_dir, method='empty')
        # Note that this ends up including all it sparents
        self.assertEqual(mapping, {
            'jquery': 'empty:',
            'underscore': 'empty:',
        })

    def test_generate_bundle_source_maps_site_empty(self):
        # This one declares exact, should still work.
        mapping = dist.generate_bundle_source_maps(
            ['site'], self.dist_dir, method='empty')
        self.assertEqual(sorted(mapping.keys()), ['jquery', 'underscore'])
        self.assertEqual(mapping['jquery'], 'empty:')
        self.assertEqual(mapping['underscore'], 'empty:')
